#!python
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True

import numpy as np
cimport numpy as cnp
cimport cython
from dipy.align.fused_types cimport floating, number
from dipy.core.interpolation cimport (_interpolate_scalar_2d,
                                      _interpolate_scalar_3d,
                                      _interpolate_vector_2d,
                                      _interpolate_vector_3d,
                                      _interpolate_scalar_nn_2d,
                                      _interpolate_scalar_nn_3d)


cdef extern from "dpy_math.h" nogil:
    double floor(double)
    double sqrt(double)
    double cos(double)
    double atan2(double, double)


def is_valid_affine(double[:, :] M, int dim):
    if M is None:
        return True
    if M.shape[0] < dim or M.shape[1] < dim + 1:
        return False
    if not np.all(np.isfinite(M)):
        return False
    return True


cdef void _compose_vector_fields_2d(floating[:, :, :] d1, floating[:, :, :] d2,
                                    double[:, :] premult_index,
                                    double[:, :] premult_disp,
                                    double time_scaling,
                                    floating[:, :, :] comp,
                                    double[:] stats) nogil:
    r"""Computes the composition of two 2D displacement fields

    Computes the composition of the two 2-D displacements d1 and d2. The
    evaluation of d2 at non-lattice points is computed using tri-linear
    interpolation. The actual composition is computed as:

    comp[i] = d1[i] + t * d2[ A * i + B * d1[i] ]

    where t = time_scaling, A = premult_index and B=premult_disp and i denotes
    the voxel coordinates of a voxel in d1's grid. Using this parameters it is
    possible to compose vector fields with arbitrary discretization: let R and
    S be the voxel-to-space transformation associated to d1 and d2,
    respectively then the composition at a voxel with coordinates i in d1's
    grid is given by:

    comp[i] = d1[i] + R*i + d2[Sinv*(R*i + d1[i])] - R*i

    (the R*i terms cancel each other) where Sinv = S^{-1}
    we can then define A = Sinv * R and B = Sinv to compute the composition
    using this function.

    Parameters
    ----------
    d1 : array, shape (R, C, 2)
        first displacement field to be applied. R, C are the number of rows
        and columns of the displacement field, respectively.
    d2 : array, shape (R', C', 2)
        second displacement field to be applied. R', C' are the number of rows
        and columns of the displacement field, respectively.
    premult_index : array, shape (3, 3)
        the matrix A in the explanation above
    premult_disp : array, shape (3, 3)
        the matrix B in the explanation above
    time_scaling : float
        this corresponds to the time scaling 't' in the above explanation
    comp : array, shape (R, C, 2), same dimension as d1
        on output, this array will contain the composition of the two fields
    stats : array, shape (3,)
        on output, this array will contain three statistics of the vector norms
        of the composition (maximum, mean, standard_deviation)

    Returns
    -------
    comp : array, shape (R, C, 2), same dimension as d1
        on output, this array will contain the composition of the two fields
    stats : array, shape (3,)
        on output, this array will contain three statistics of the vector norms
        of the composition (maximum, mean, standard_deviation)

    Notes
    -----
    If d1[r,c] lies outside the domain of d2, then comp[r,c] will contain a
    zero vector.

    Warning: it is possible to use the same array reference for d1 and comp to
    effectively update d1 to the composition of d1 and d2 because previously
    updated values from d1 are no longer used (this is done to save memory and
    time). However, using the same array for d2 and comp may not be the
    intended operation (see comment below).

    """
    cdef:
        cnp.npy_intp nr1 = d1.shape[0]
        cnp.npy_intp nc1 = d1.shape[1]
        cnp.npy_intp nr2 = d2.shape[0]
        cnp.npy_intp nc2 = d2.shape[1]
        int inside, cnt = 0
        double maxNorm = 0
        double meanNorm = 0
        double stdNorm = 0
        double nn
        cnp.npy_intp i, j
        double di, dj, dii, djj, diii, djjj

    for i in range(nr1):
        for j in range(nc1):

            # This is the only place we access d1[i, j]
            dii = d1[i, j, 0]
            djj = d1[i, j, 1]

            if premult_disp is None:
                di = dii
                dj = djj
            else:
                di = _apply_affine_2d_x0(dii, djj, 0, premult_disp)
                dj = _apply_affine_2d_x1(dii, djj, 0, premult_disp)

            if premult_index is None:
                diii = i
                djjj = j
            else:
                diii = _apply_affine_2d_x0(i, j, 1, premult_index)
                djjj = _apply_affine_2d_x1(i, j, 1, premult_index)

            diii += di
            djjj += dj

            # If d1 and comp are the same array, this will correctly update
            # d1[i,j], which will never be accessed again
            # If d2 and comp are the same array, then (diii, djjj) may be
            # in the neighborhood of a previously updated vector from d2,
            # which may be problematic
            inside = _interpolate_vector_2d[floating](d2, diii, djjj,
                                                      &comp[i, j, 0])

            if inside == 1:
                comp[i, j, 0] = time_scaling * comp[i, j, 0] + dii
                comp[i, j, 1] = time_scaling * comp[i, j, 1] + djj
                nn = comp[i, j, 0] ** 2 + comp[i, j, 1] ** 2
                meanNorm += nn
                stdNorm += nn * nn
                cnt += 1
                if(maxNorm < nn):
                    maxNorm = nn
            else:
                comp[i, j, 0] = 0
                comp[i, j, 1] = 0

    meanNorm /= cnt
    stats[0] = sqrt(maxNorm)
    stats[1] = sqrt(meanNorm)
    stats[2] = sqrt(stdNorm / cnt - meanNorm * meanNorm)


def compose_vector_fields_2d(floating[:, :, :] d1, floating[:, :, :] d2,
                             double[:, :] premult_index,
                             double[:, :] premult_disp,
                             double time_scaling,
                             floating[:, :, :] comp):
    r"""Computes the composition of two 2D displacement fields

    Computes the composition of the two 2-D displacements d1 and d2. The
    evaluation of d2 at non-lattice points is computed using tri-linear
    interpolation. The actual composition is computed as:

    comp[i] = d1[i] + t * d2[ A * i + B * d1[i] ]

    where t = time_scaling, A = premult_index and B=premult_disp and i denotes
    the voxel coordinates of a voxel in d1's grid. Using this parameters it is
    possible to compose vector fields with arbitrary discretizations: let R and
    S be the voxel-to-space transformation associated to d1 and d2,
    respectively then the composition at a voxel with coordinates i in d1's
    grid is given by:

    comp[i] = d1[i] + R*i + d2[Sinv*(R*i + d1[i])] - R*i

    (the R*i terms cancel each other) where Sinv = S^{-1}
    we can then define A = Sinv * R and B = Sinv to compute the composition
    using this function.

    Parameters
    ----------
    d1 : array, shape (R, C, 2)
        first displacement field to be applied. R, C are the number of rows
        and columns of the displacement field, respectively.
    d2 : array, shape (R', C', 2)
        second displacement field to be applied. R', C' are the number of rows
        and columns of the displacement field, respectively.
    premult_index : array, shape (3, 3)
        the matrix A in the explanation above
    premult_disp : array, shape (3, 3)
        the matrix B in the explanation above
    time_scaling : float
        this corresponds to the time scaling 't' in the above explanation
    comp : array, shape (R, C, 2)
        the buffer to write the composition to. If None, the buffer is created
        internally

    Returns
    -------
    comp : array, shape (R, C, 2), same dimension as d1
        on output, this array will contain the composition of the two fields
    stats : array, shape (3,)
        on output, this array will contain three statistics of the vector norms
        of the composition (maximum, mean, standard_deviation)
    """
    cdef:
        double[:] stats = np.zeros(shape=(3,), dtype=np.float64)

    if comp is None:
        comp = np.zeros_like(d1)

    if not is_valid_affine(premult_index, 2):
        raise ValueError("Invalid index multiplication matrix")
    if not is_valid_affine(premult_disp, 2):
        raise ValueError("Invalid displacement multiplication matrix")

    _compose_vector_fields_2d[floating](d1, d2, premult_index, premult_disp,
                                        time_scaling, comp, stats)
    return np.asarray(comp), np.asarray(stats)


cdef void _compose_vector_fields_3d(floating[:, :, :, :] d1,
                                    floating[:, :, :, :] d2,
                                    double[:, :] premult_index,
                                    double[:, :] premult_disp,
                                    double t,
                                    floating[:, :, :, :] comp,
                                    double[:] stats) nogil:
    r"""Computes the composition of two 3D displacement fields

    Computes the composition of the two 3-D displacements d1 and d2. The
    evaluation of d2 at non-lattice points is computed using tri-linear
    interpolation. The actual composition is computed as:

    comp[i] = d1[i] + t * d2[ A * i + B * d1[i] ]

    where t = time_scaling, A = premult_index and B=premult_disp and i denotes
    the voxel coordinates of a voxel in d1's grid. Using this parameters it is
    possible to compose vector fields with arbitrary discretization: let R and
    S be the voxel-to-space transformation associated to d1 and d2,
    respectively then the composition at a voxel with coordinates i in d1's
    grid is given by:

    comp[i] = d1[i] + R*i + d2[Sinv*(R*i + d1[i])] - R*i

    (the R*i terms cancel each other) where Sinv = S^{-1}
    we can then define A = Sinv * R and B = Sinv to compute the composition
    using this function.

    Parameters
    ----------
    d1 : array, shape (S, R, C, 3)
        first displacement field to be applied. S, R, C are the number of
        slices, rows and columns of the displacement field, respectively.
    d2 : array, shape (S', R', C', 3)
        second displacement field to be applied. R', C' are the number of rows
        and columns of the displacement field, respectively.
    premult_index : array, shape (4, 4)
        the matrix A in the explanation above
    premult_disp : array, shape (4, 4)
        the matrix B in the explanation above
    time_scaling : float
        this corresponds to the time scaling 't' in the above explanation
    comp : array, shape (S, R, C, 3), same dimension as d1
        on output, this array will contain the composition of the two fields
    stats : array, shape (3,)
        on output, this array will contain three statistics of the vector norms
        of the composition (maximum, mean, standard_deviation)

    Returns
    -------
    comp : array, shape (S, R, C, 3), same dimension as d1
        on output, this array will contain the composition of the two fields
    stats : array, shape (3,)
        on output, this array will contain three statistics of the vector norms
        of the composition (maximum, mean, standard_deviation)

    Notes
    -----
    If d1[s,r,c] lies outside the domain of d2, then comp[s,r,c] will contain
    a zero vector.

    Warning: it is possible to use the same array reference for d1 and comp to
    effectively update d1 to the composition of d1 and d2 because previously
    updated values from d1 are no longer used (this is done to save memory and
    time). However, using the same array for d2 and comp may not be the
    intended operation (see comment below).
    """
    cdef:
        cnp.npy_intp ns1 = d1.shape[0]
        cnp.npy_intp nr1 = d1.shape[1]
        cnp.npy_intp nc1 = d1.shape[2]
        cnp.npy_intp ns2 = d2.shape[0]
        cnp.npy_intp nr2 = d2.shape[1]
        cnp.npy_intp nc2 = d2.shape[2]
        int inside, cnt = 0
        double maxNorm = 0
        double meanNorm = 0
        double stdNorm = 0
        double nn
        cnp.npy_intp i, j, k
        double di, dj, dk, dii, djj, dkk, diii, djjj, dkkk
    for k in range(ns1):
        for i in range(nr1):
            for j in range(nc1):

                # This is the only place we access d1[k, i, j]
                dkk = d1[k, i, j, 0]
                dii = d1[k, i, j, 1]
                djj = d1[k, i, j, 2]

                if premult_disp is None:
                    dk = dkk
                    di = dii
                    dj = djj
                else:
                    dk = _apply_affine_3d_x0(dkk, dii, djj, 0, premult_disp)
                    di = _apply_affine_3d_x1(dkk, dii, djj, 0, premult_disp)
                    dj = _apply_affine_3d_x2(dkk, dii, djj, 0, premult_disp)

                if premult_index is None:
                    dkkk = k
                    diii = i
                    djjj = j
                else:
                    dkkk = _apply_affine_3d_x0(k, i, j, 1, premult_index)
                    diii = _apply_affine_3d_x1(k, i, j, 1, premult_index)
                    djjj = _apply_affine_3d_x2(k, i, j, 1, premult_index)

                dkkk += dk
                diii += di
                djjj += dj

                # If d1 and comp are the same array, this will correctly update
                # d1[k,i,j], which will never be accessed again
                # If d2 and comp are the same array, then (dkkk, diii, djjj)
                # may be in the neighborhood of a previously updated vector
                # from d2, which may be problematic
                inside = _interpolate_vector_3d[floating](d2, dkkk, diii, djjj,
                                                          &comp[k, i, j, 0])

                if inside == 1:
                    comp[k, i, j, 0] = t * comp[k, i, j, 0] + dkk
                    comp[k, i, j, 1] = t * comp[k, i, j, 1] + dii
                    comp[k, i, j, 2] = t * comp[k, i, j, 2] + djj
                    nn = (comp[k, i, j, 0] ** 2 + comp[k, i, j, 1] ** 2 +
                          comp[k, i, j, 2]**2)
                    meanNorm += nn
                    stdNorm += nn * nn
                    cnt += 1
                    if(maxNorm < nn):
                        maxNorm = nn
                else:
                    comp[k, i, j, 0] = 0
                    comp[k, i, j, 1] = 0
                    comp[k, i, j, 2] = 0
    meanNorm /= cnt
    stats[0] = sqrt(maxNorm)
    stats[1] = sqrt(meanNorm)
    stats[2] = sqrt(stdNorm / cnt - meanNorm * meanNorm)


def compose_vector_fields_3d(floating[:, :, :, :] d1, floating[:, :, :, :] d2,
                             double[:, :] premult_index,
                             double[:, :] premult_disp,
                             double time_scaling,
                             floating[:, :, :, :] comp):
    r"""Computes the composition of two 3D displacement fields

    Computes the composition of the two 3-D displacements d1 and d2. The
    evaluation of d2 at non-lattice points is computed using tri-linear
    interpolation. The actual composition is computed as:

    comp[i] = d1[i] + t * d2[ A * i + B * d1[i] ]

    where t = time_scaling, A = premult_index and B=premult_disp and i denotes
    the voxel coordinates of a voxel in d1's grid. Using this parameters it is
    possible to compose vector fields with arbitrary discretization: let R and
    S be the voxel-to-space transformation associated to d1 and d2,
    respectively then the composition at a voxel with coordinates i in d1's
    grid is given by:

    comp[i] = d1[i] + R*i + d2[Sinv*(R*i + d1[i])] - R*i

    (the R*i terms cancel each other) where Sinv = S^{-1}
    we can then define A = Sinv * R and B = Sinv to compute the composition
    using this function.

    Parameters
    ----------
    d1 : array, shape (S, R, C, 3)
        first displacement field to be applied. S, R, C are the number of
        slices, rows and columns of the displacement field, respectively.
    d2 : array, shape (S', R', C', 3)
        second displacement field to be applied. R', C' are the number of rows
        and columns of the displacement field, respectively.
    premult_index : array, shape (4, 4)
        the matrix A in the explanation above
    premult_disp : array, shape (4, 4)
        the matrix B in the explanation above
    time_scaling : float
        this corresponds to the time scaling 't' in the above explanation
    comp : array, shape (S, R, C, 3), same dimension as d1
        the buffer to write the composition to. If None, the buffer will be
        created internally

    Returns
    -------
    comp : array, shape (S, R, C, 3), same dimension as d1
        on output, this array will contain the composition of the two fields
    stats : array, shape (3,)
        on output, this array will contain three statistics of the vector norms
        of the composition (maximum, mean, standard_deviation)

    Notes
    -----
    If d1[s,r,c] lies outside the domain of d2, then comp[s,r,c] will contain
    a zero vector.
    """
    cdef:
        double[:] stats = np.zeros(shape=(3,), dtype=np.float64)

    if comp is None:
        comp = np.zeros_like(d1)

    if not is_valid_affine(premult_index, 3):
        raise ValueError("Invalid index pre-multiplication matrix")
    if not is_valid_affine(premult_disp, 3):
        raise ValueError("Invalid displacement pre-multiplication matrix")

    _compose_vector_fields_3d[floating](d1, d2, premult_index, premult_disp,
                                        time_scaling, comp, stats)
    return np.asarray(comp), np.asarray(stats)


def invert_vector_field_fixed_point_2d(floating[:, :, :] d,
                                       double[:, :] d_world2grid,
                                       double[:] spacing,
                                       int max_iter, double tolerance,
                                       floating[:, :, :] start=None):
    r"""Computes the inverse of a 2D displacement fields

    Computes the inverse of the given 2-D displacement field d using the
    fixed-point algorithm [1].

    [1] Chen, M., Lu, W., Chen, Q., Ruchala, K. J., & Olivera, G. H. (2008).
        A simple fixed-point approach to invert a deformation field.
        Medical Physics, 35(1), 81. doi:10.1118/1.2816107

    Parameters
    ----------
    d : array, shape (R, C, 2)
        the 2-D displacement field to be inverted
    d_world2grid : array, shape (3, 3)
        the space-to-grid transformation associated to the displacement field
        d (transforming physical space coordinates to voxel coordinates of the
        displacement field grid)
    spacing :array, shape (2,)
        the spacing between voxels (voxel size along each axis)
    max_iter : int
        maximum number of iterations to be performed
    tolerance : float
        maximum tolerated inversion error
    start : array, shape (R, C)
        an approximation to the inverse displacement field (if no approximation
        is available, None can be provided and the start displacement field
        will be zero)

    Returns
    -------
    p : array, shape (R, C, 2)
        the inverse displacement field

    Notes
    -----
    We assume that the displacement field is an endomorphism so that the shape
    and voxel-to-space transformation of the inverse field's discretization is
    the same as those of the input displacement field. The 'inversion error' at
    iteration t is defined as the mean norm of the displacement vectors of the
    input displacement field composed with the inverse at iteration t.
    """
    cdef:
        cnp.npy_intp nr = d.shape[0]
        cnp.npy_intp nc = d.shape[1]
        int iter_count, current, flag
        double difmag, mag, maxlen, step_factor
        double epsilon
        double error = 1 + tolerance
        double di, dj, dii, djj
        double sr = spacing[0], sc = spacing[1]

    ftype = np.asarray(d).dtype
    cdef:
        double[:] stats = np.zeros(shape=(2,), dtype=np.float64)
        double[:] substats = np.empty(shape=(3,), dtype=np.float64)
        double[:, :] norms = np.zeros(shape=(nr, nc), dtype=np.float64)
        floating[:, :, :] p = np.zeros(shape=(nr, nc, 2), dtype=ftype)
        floating[:, :, :] q = np.zeros(shape=(nr, nc, 2), dtype=ftype)

    if not is_valid_affine(d_world2grid, 2):
        raise ValueError("Invalid world-to-image transform")

    if start is not None:
        p[...] = start

    with nogil:
        iter_count = 0
        while (iter_count < max_iter) and (tolerance < error):
            if iter_count == 0:
                epsilon = 0.75
            else:
                epsilon = 0.5
            _compose_vector_fields_2d[floating](p, d, None, d_world2grid,
                                                1.0, q, substats)
            difmag = 0
            error = 0
            for i in range(nr):
                for j in range(nc):
                    mag = sqrt((q[i, j, 0]/sr) ** 2 + (q[i, j, 1]/sc) ** 2)
                    norms[i, j] = mag
                    error += mag
                    if(difmag < mag):
                        difmag = mag
            maxlen = difmag * epsilon
            for i in range(nr):
                for j in range(nc):
                    if norms[i, j] > maxlen:
                        step_factor = epsilon * maxlen / norms[i, j]
                    else:
                        step_factor = epsilon
                    p[i, j, 0] = p[i, j, 0] - step_factor * q[i, j, 0]
                    p[i, j, 1] = p[i, j, 1] - step_factor * q[i, j, 1]
            error /= (nr * nc)
            iter_count += 1
        stats[0] = substats[1]
        stats[1] = iter_count
    return np.asarray(p)


def invert_vector_field_fixed_point_3d(floating[:, :, :, :] d,
                                       double[:, :] d_world2grid,
                                       double[:] spacing,
                                       int max_iter, double tol,
                                       floating[:, :, :, :] start=None):
    r"""Computes the inverse of a 3D displacement fields

    Computes the inverse of the given 3-D displacement field d using the
    fixed-point algorithm [1].

    [1] Chen, M., Lu, W., Chen, Q., Ruchala, K. J., & Olivera, G. H. (2008).
        A simple fixed-point approach to invert a deformation field.
        Medical Physics, 35(1), 81. doi:10.1118/1.2816107

    Parameters
    ----------
    d : array, shape (S, R, C, 3)
        the 3-D displacement field to be inverted
    d_world2grid : array, shape (4, 4)
        the space-to-grid transformation associated to the displacement field
        d (transforming physical space coordinates to voxel coordinates of the
        displacement field grid)
    spacing :array, shape (3,)
        the spacing between voxels (voxel size along each axis)
    max_iter : int
        maximum number of iterations to be performed
    tol : float
        maximum tolerated inversion error
    start : array, shape (S, R, C)
        an approximation to the inverse displacement field (if no approximation
        is available, None can be provided and the start displacement field
        will be zero)

    Returns
    -------
    p : array, shape (S, R, C, 3)
        the inverse displacement field

    Notes
    -----
    We assume that the displacement field is an endomorphism so that the shape
    and voxel-to-space transformation of the inverse field's discretization is
    the same as those of the input displacement field. The 'inversion error' at
    iteration t is defined as the mean norm of the displacement vectors of the
    input displacement field composed with the inverse at iteration t.
    """
    cdef:
        cnp.npy_intp ns = d.shape[0]
        cnp.npy_intp nr = d.shape[1]
        cnp.npy_intp nc = d.shape[2]
        int iter_count, current
        double dkk, dii, djj, dk, di, dj
        double difmag, mag, maxlen, step_factor
        double epsilon = 0.5
        double error = 1 + tol
        double ss = spacing[0], sr = spacing[1], sc = spacing[2]

    ftype = np.asarray(d).dtype
    cdef:
        double[:] stats = np.zeros(shape=(2,), dtype=np.float64)
        double[:] substats = np.zeros(shape=(3,), dtype=np.float64)
        double[:, :, :] norms = np.zeros(shape=(ns, nr, nc), dtype=np.float64)
        floating[:, :, :, :] p = np.zeros(shape=(ns, nr, nc, 3), dtype=ftype)
        floating[:, :, :, :] q = np.zeros(shape=(ns, nr, nc, 3), dtype=ftype)

    if not is_valid_affine(d_world2grid, 3):
        raise ValueError("Invalid world-to-image transform")

    if start is not None:
        p[...] = start

    with nogil:
        iter_count = 0
        difmag = 1
        while (0.1 < difmag) and (iter_count < max_iter) and (tol < error):
            if iter_count == 0:
                epsilon = 0.75
            else:
                epsilon = 0.5
            _compose_vector_fields_3d[floating](p, d, None, d_world2grid,
                                                1.0, q, substats)
            difmag = 0
            error = 0
            for k in range(ns):
                for i in range(nr):
                    for j in range(nc):
                        mag = sqrt((q[k, i, j, 0]/ss) ** 2 +
                                   (q[k, i, j, 1]/sr) ** 2 +
                                   (q[k, i, j, 2]/sc) ** 2)
                        norms[k, i, j] = mag
                        error += mag
                        if(difmag < mag):
                            difmag = mag
            maxlen = difmag*epsilon
            for k in range(ns):
                for i in range(nr):
                    for j in range(nc):
                        if norms[k, i, j] > maxlen:
                            step_factor = epsilon * maxlen / norms[k, i, j]
                        else:
                            step_factor = epsilon
                        p[k, i, j, 0] = (p[k, i, j, 0] -
                                         step_factor * q[k, i, j, 0])
                        p[k, i, j, 1] = (p[k, i, j, 1] -
                                         step_factor * q[k, i, j, 1])
                        p[k, i, j, 2] = (p[k, i, j, 2] -
                                         step_factor * q[k, i, j, 2])
            error /= (ns * nr * nc)
            iter_count += 1
        stats[0] = error
        stats[1] = iter_count
    return np.asarray(p)


def simplify_warp_function_2d(floating[:, :, :] d,
                              double[:, :] affine_idx_in,
                              double[:, :] affine_idx_out,
                              double[:, :] affine_disp,
                              int[:] out_shape):
    r"""
    Simplifies a nonlinear warping function combined with an affine transform

    Modifies the given deformation field by incorporating into it a
    an affine transformation and voxel-to-space transforms associated to
    the discretization of its domain and codomain.
    The resulting transformation may be regarded as operating on the
    image spaces given by the domain and codomain discretization.
    More precisely, the resulting transform is of the form:

    (1) T[i] = W * d[U * i] + V * i

    Where U = affine_idx_in, V = affine_idx_out, W = affine_disp.

    Parameters
    ----------
    d : array, shape (R', C', 2)
        the non-linear part of the transformation (displacement field)
    affine_idx_in : array, shape (3, 3)
        the matrix U in eq. (1) above
    affine_idx_out : array, shape (3, 3)
        the matrix V in eq. (1) above
    affine_disp : array, shape (3, 3)
        the matrix W in eq. (1) above
    out_shape : array, shape (2,)
        the number of rows and columns of the sampling grid

    Returns
    -------
    out : array, shape = out_shape
        the deformation field `out` associated with `T` in eq. (1) such that:
        T[i] = i + out[i]

    Notes
    -----
    Both the direct and inverse transforms of a DiffeomorphicMap can be written
    in this form:

    Direct:  Let D be the voxel-to-space transform of the domain's
             discretization, P be the pre-align matrix, Rinv the space-to-voxel
             transform of the reference grid (the grid the displacement field
             is defined on) and Cinv be the space-to-voxel transform of the
             codomain's discretization. Then, for each i in the domain's grid,
             the direct transform is given by

             (2) T[i] = Cinv * d[Rinv * P * D * i] + Cinv * P * D * i

             and we identify U = Rinv * P * D, V = Cinv * P * D, W = Cinv

    Inverse: Let C be the voxel-to-space transform of the codomain's
             discretization, Pinv be the inverse of the pre-align matrix, Rinv
             the space-to-voxel transform of the reference grid (the grid the
             displacement field is defined on) and Dinv be the space-to-voxel
             transform of the domain's discretization. Then, for each j in the
             codomain's grid, the inverse transform is given by

             (3) Tinv[j] = Dinv * Pinv * d[Rinv * C * j] + Dinv * Pinv * C * j

             and we identify U = Rinv * C, V = Dinv * Pinv * C, W = Dinv * Pinv
    """
    cdef:
        cnp.npy_intp nrows = out_shape[0]
        cnp.npy_intp ncols = out_shape[1]
        cnp.npy_intp i, j
        double di, dj, dii, djj
        floating[:] tmp = np.zeros((2,), dtype=np.asarray(d).dtype)
        floating[:, :, :] out = np.zeros(shape=(nrows, ncols, 2),
                                         dtype=np.asarray(d).dtype)

    if not is_valid_affine(affine_idx_in, 2):
        raise ValueError("Invalid inner index multiplication matrix")
    if not is_valid_affine(affine_idx_out, 2):
        raise ValueError("Invalid outer index multiplication matrix")
    if not is_valid_affine(affine_disp, 2):
        raise ValueError("Invalid displacement multiplication matrix")

    with nogil:

        for i in range(nrows):
            for j in range(ncols):
                # Apply inner index pre-multiplication
                if affine_idx_in is None:
                    dii = d[i, j, 0]
                    djj = d[i, j, 1]
                else:
                    di = _apply_affine_2d_x0(
                        i, j, 1, affine_idx_in)
                    dj = _apply_affine_2d_x1(
                        i, j, 1, affine_idx_in)
                    _interpolate_vector_2d[floating](d, di, dj, &tmp[0])
                    dii = tmp[0]
                    djj = tmp[1]

                # Apply displacement multiplication
                if affine_disp is not None:
                    di = _apply_affine_2d_x0(
                        dii, djj, 0, affine_disp)
                    dj = _apply_affine_2d_x1(
                        dii, djj, 0, affine_disp)
                else:
                    di = dii
                    dj = djj

                # Apply outer index multiplication and add the displacements
                if affine_idx_out is not None:
                    out[i, j, 0] = di + _apply_affine_2d_x0(i, j, 1,
                                                            affine_idx_out) - i
                    out[i, j, 1] = dj + _apply_affine_2d_x1(i, j, 1,
                                                            affine_idx_out) - j
                else:
                    out[i, j, 0] = di
                    out[i, j, 1] = dj
    return np.asarray(out)


def simplify_warp_function_3d(floating[:, :, :, :] d,
                              double[:, :] affine_idx_in,
                              double[:, :] affine_idx_out,
                              double[:, :] affine_disp,
                              int[:] out_shape):
    r"""
    Simplifies a nonlinear warping function combined with an affine transform

    Modifies the given deformation field by incorporating into it
    an affine transformation and voxel-to-space transforms associated with
    the discretization of its domain and codomain.

    The resulting transformation may be regarded as operating on the
    image spaces given by the domain and codomain discretization.
    More precisely, the resulting transform is of the form:

    (1) T[i] = W * d[U * i] + V * i

    Where U = affine_idx_in, V = affine_idx_out, W = affine_disp.

    Parameters
    ----------
    d : array, shape (S', R', C', 3)
        the non-linear part of the transformation (displacement field)
    affine_idx_in : array, shape (4, 4)
        the matrix U in eq. (1) above
    affine_idx_out : array, shape (4, 4)
        the matrix V in eq. (1) above
    affine_disp : array, shape (4, 4)
        the matrix W in eq. (1) above
    out_shape : array, shape (3,)
        the number of slices, rows and columns of the sampling grid

    Returns
    -------
    out : array, shape = out_shape
        the deformation field `out` associated with `T` in eq. (1) such that:
        T[i] = i + out[i]

    Notes
    -----
    Both the direct and inverse transforms of a DiffeomorphicMap can be written
    in this form:

    Direct:  Let D be the voxel-to-space transform of the domain's
             discretization, P be the pre-align matrix, Rinv the space-to-voxel
             transform of the reference grid (the grid the displacement field
             is defined on) and Cinv be the space-to-voxel transform of the
             codomain's discretization. Then, for each i in the domain's grid,
             the direct transform is given by

             (2) T[i] = Cinv * d[Rinv * P * D * i] + Cinv * P * D * i

             and we identify U = Rinv * P * D, V = Cinv * P * D, W = Cinv

    Inverse: Let C be the voxel-to-space transform of the codomain's
             discretization, Pinv be the inverse of the pre-align matrix, Rinv
             the space-to-voxel transform of the reference grid (the grid the
             displacement field is defined on) and Dinv be the space-to-voxel
             transform of the domain's discretization. Then, for each j in the
             codomain's grid, the inverse transform is given by

             (3) Tinv[j] = Dinv * Pinv * d[Rinv * C * j] + Dinv * Pinv * C * j

             and we identify U = Rinv * C, V = Dinv * Pinv * C, W = Dinv * Pinv

    """
    cdef:
        cnp.npy_intp nslices = out_shape[0]
        cnp.npy_intp nrows = out_shape[1]
        cnp.npy_intp ncols = out_shape[2]
        cnp.npy_intp i, j, k, inside
        double di, dj, dk, dii, djj, dkk
        floating[:] tmp = np.zeros((3,), dtype=np.asarray(d).dtype)
        floating[:, :, :, :] out = np.zeros(shape=(nslices, nrows, ncols, 3),
                                            dtype=np.asarray(d).dtype)

    if not is_valid_affine(affine_idx_in, 3):
        raise ValueError("Invalid inner index multiplication matrix")
    if not is_valid_affine(affine_idx_out, 3):
        raise ValueError("Invalid outer index multiplication matrix")
    if not is_valid_affine(affine_disp, 3):
        raise ValueError("Invalid displacement multiplication matrix")

    with nogil:

        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    if affine_idx_in is None:
                        dkk = d[k, i, j, 0]
                        dii = d[k, i, j, 1]
                        djj = d[k, i, j, 2]
                    else:
                        dk = _apply_affine_3d_x0(
                            k, i, j, 1, affine_idx_in)
                        di = _apply_affine_3d_x1(
                            k, i, j, 1, affine_idx_in)
                        dj = _apply_affine_3d_x2(
                            k, i, j, 1, affine_idx_in)
                        inside = _interpolate_vector_3d[floating](d, dk, di,
                                                                  dj, &tmp[0])
                        dkk = tmp[0]
                        dii = tmp[1]
                        djj = tmp[2]

                    if affine_disp is not None:
                        dk = _apply_affine_3d_x0(
                            dkk, dii, djj, 0, affine_disp)
                        di = _apply_affine_3d_x1(
                            dkk, dii, djj, 0, affine_disp)
                        dj = _apply_affine_3d_x2(
                            dkk, dii, djj, 0, affine_disp)
                    else:
                        dk = dkk
                        di = dii
                        dj = djj

                    if affine_idx_out is not None:
                        out[k, i, j, 0] = dk +\
                            _apply_affine_3d_x0(k, i, j, 1, affine_idx_out) - k
                        out[k, i, j, 1] = di +\
                            _apply_affine_3d_x1(k, i, j, 1, affine_idx_out) - i
                        out[k, i, j, 2] = dj +\
                            _apply_affine_3d_x2(k, i, j, 1, affine_idx_out) - j
                    else:
                        out[k, i, j, 0] = dk
                        out[k, i, j, 1] = di
                        out[k, i, j, 2] = dj
    return np.asarray(out)


def reorient_vector_field_2d(floating[:, :, :] d,
                             double[:, :] affine):
    r"""Linearly transforms all vectors of a 2D displacement field

    Modifies the input displacement field by multiplying each displacement
    vector by the given matrix. Note that the elements of the displacement
    field are vectors, not points, so their last homogeneous coordinate is
    zero, not one, and therefore the translation component of the affine
    transform will not have any effect on them.

    Parameters
    ----------
    d : array, shape (R, C, 2)
        the displacement field to be re-oriented
    affine: array, shape (3, 3)
        the matrix to be applied
    """
    cdef:
        cnp.npy_intp nrows = d.shape[0]
        cnp.npy_intp ncols = d.shape[1]
        cnp.npy_intp i, j
        double di, dj

    if not is_valid_affine(affine, 2):
        raise ValueError("Invalid affine transform matrix")

    if affine is None:
        return

    with nogil:
        for i in range(nrows):
            for j in range(ncols):
                di = d[i, j, 0]
                dj = d[i, j, 1]
                d[i, j, 0] = _apply_affine_2d_x0(di, dj, 0, affine)
                d[i, j, 1] = _apply_affine_2d_x1(di, dj, 0, affine)


def reorient_vector_field_3d(floating[:, :, :, :] d,
                             double[:, :] affine):
    r"""Linearly transforms all vectors of a 3D displacement field

    Modifies the input displacement field by multiplying each displacement
    vector by the given matrix. Note that the elements of the displacement
    field are vectors, not points, so their last homogeneous coordinate is
    zero, not one, and therefore the translation component of the affine
    transform will not have any effect on them.

    Parameters
    ----------
    d : array, shape (S, R, C, 3)
        the displacement field to be re-oriented
    affine : array, shape (4, 4)
        the matrix to be applied
    """
    cdef:
        cnp.npy_intp nslices = d.shape[0]
        cnp.npy_intp nrows = d.shape[1]
        cnp.npy_intp ncols = d.shape[2]
        cnp.npy_intp i, j, k
        double di, dj, dk

    if not is_valid_affine(affine, 3):
        raise ValueError("Invalid affine transform matrix")

    if affine is None:
        return

    with nogil:
        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    dk = d[k, i, j, 0]
                    di = d[k, i, j, 1]
                    dj = d[k, i, j, 2]
                    d[k, i, j, 0] = _apply_affine_3d_x0(dk, di, dj, 0, affine)
                    d[k, i, j, 1] = _apply_affine_3d_x1(dk, di, dj, 0, affine)
                    d[k, i, j, 2] = _apply_affine_3d_x2(dk, di, dj, 0, affine)


def downsample_scalar_field_3d(floating[:, :, :] field):
    r"""Down-samples the input volume by a factor of 2

    Down-samples the input volume by a factor of 2. The value at each voxel
    of the resulting volume is the average of its surrounding voxels in the
    original volume.

    Parameters
    ----------
    field : array, shape (S, R, C)
        the volume to be down-sampled

    Returns
    -------
    down : array, shape (S', R', C')
        the down-sampled displacement field, where S' = ceil(S/2),
        R'= ceil(R/2), C'=ceil(C/2)
    """
    ftype = np.asarray(field).dtype
    cdef:
        cnp.npy_intp ns = field.shape[0]
        cnp.npy_intp nr = field.shape[1]
        cnp.npy_intp nc = field.shape[2]
        cnp.npy_intp nns = (ns + 1) // 2
        cnp.npy_intp nnr = (nr + 1) // 2
        cnp.npy_intp nnc = (nc + 1) // 2
        cnp.npy_intp i, j, k, ii, jj, kk
        floating[:, :, :] down = np.zeros((nns, nnr, nnc), dtype=ftype)
        int[:, :, :] cnt = np.zeros((nns, nnr, nnc), dtype=np.int32)

    with nogil:
        for k in range(ns):
            for i in range(nr):
                for j in range(nc):
                    kk = k // 2
                    ii = i // 2
                    jj = j // 2
                    down[kk, ii, jj] += field[k, i, j]
                    cnt[kk, ii, jj] += 1
        for k in range(nns):
            for i in range(nnr):
                for j in range(nnc):
                    if cnt[k, i, j] > 0:
                        down[k, i, j] /= cnt[k, i, j]
    return np.asarray(down)


def downsample_displacement_field_3d(floating[:, :, :, :] field):
    r"""Down-samples the input 3D vector field by a factor of 2

    Down-samples the input vector field by a factor of 2. This operation
    is equivalent to dividing the input image into 2x2x2 cubes and averaging
    the 8 vectors. The resulting field consists of these average vectors.

    Parameters
    ----------
    field : array, shape (S, R, C)
        the vector field to be down-sampled

    Returns
    -------
    down : array, shape (S', R', C')
        the down-sampled displacement field, where S' = ceil(S/2),
        R'= ceil(R/2), C'=ceil(C/2)
    """
    ftype = np.asarray(field).dtype
    cdef:
        cnp.npy_intp ns = field.shape[0]
        cnp.npy_intp nr = field.shape[1]
        cnp.npy_intp nc = field.shape[2]
        cnp.npy_intp nns = (ns + 1) // 2
        cnp.npy_intp nnr = (nr + 1) // 2
        cnp.npy_intp nnc = (nc + 1) // 2
        cnp.npy_intp i, j, k, ii, jj, kk
        floating[:, :, :, :] down = np.zeros((nns, nnr, nnc, 3), dtype=ftype)
        int[:, :, :] cnt = np.zeros((nns, nnr, nnc), dtype=np.int32)

    with nogil:

        for k in range(ns):
            for i in range(nr):
                for j in range(nc):
                    kk = k // 2
                    ii = i // 2
                    jj = j // 2
                    down[kk, ii, jj, 0] += field[k, i, j, 0]
                    down[kk, ii, jj, 1] += field[k, i, j, 1]
                    down[kk, ii, jj, 2] += field[k, i, j, 2]
                    cnt[kk, ii, jj] += 1
        for k in range(nns):
            for i in range(nnr):
                for j in range(nnc):
                    if cnt[k, i, j] > 0:
                        down[k, i, j, 0] /= cnt[k, i, j]
                        down[k, i, j, 1] /= cnt[k, i, j]
                        down[k, i, j, 2] /= cnt[k, i, j]
    return np.asarray(down)


def downsample_scalar_field_2d(floating[:, :] field):
    r"""Down-samples the input 2D image by a factor of 2

    Down-samples the input image by a factor of 2. The value at each pixel
    of the resulting image is the average of its surrounding pixels in the
    original image.

    Parameters
    ----------
    field : array, shape (R, C)
        the image to be down-sampled

    Returns
    -------
    down : array, shape (R', C')
        the down-sampled displacement field, where R'= ceil(R/2), C'=ceil(C/2)
    """
    ftype = np.asarray(field).dtype
    cdef:
        cnp.npy_intp nr = field.shape[0]
        cnp.npy_intp nc = field.shape[1]
        cnp.npy_intp nnr = (nr + 1) // 2
        cnp.npy_intp nnc = (nc + 1) // 2
        cnp.npy_intp i, j, ii, jj
        floating[:, :] down = np.zeros(shape=(nnr, nnc), dtype=ftype)
        int[:, :] cnt = np.zeros(shape=(nnr, nnc), dtype=np.int32)
    with nogil:

        for i in range(nr):
            for j in range(nc):
                ii = i // 2
                jj = j // 2
                down[ii, jj] += field[i, j]
                cnt[ii, jj] += 1
        for i in range(nnr):
            for j in range(nnc):
                if cnt[i, j] > 0:
                    down[i, j] /= cnt[i, j]
    return np.asarray(down)


def downsample_displacement_field_2d(floating[:, :, :] field):
    r"""Down-samples the 2D input vector field by a factor of 2
    Down-samples the input vector field by a factor of 2. The value at each
    pixel of the resulting field is the average of its surrounding pixels in
    the original field.

    Parameters
    ----------
    field : array, shape (R, C)
        the vector field to be down-sampled

    Returns
    -------
    down : array, shape (R', C')
        the down-sampled displacement field, where R'= ceil(R/2), C'=ceil(C/2),
    """
    ftype = np.asarray(field).dtype
    cdef:
        cnp.npy_intp nr = field.shape[0]
        cnp.npy_intp nc = field.shape[1]
        cnp.npy_intp nnr = (nr + 1) // 2
        cnp.npy_intp nnc = (nc + 1) // 2
        cnp.npy_intp i, j, ii, jj
        floating[:, :, :] down = np.zeros((nnr, nnc, 2), dtype=ftype)
        int[:, :] cnt = np.zeros((nnr, nnc), dtype=np.int32)

    with nogil:

        for i in range(nr):
            for j in range(nc):
                ii = i // 2
                jj = j // 2
                down[ii, jj, 0] += field[i, j, 0]
                down[ii, jj, 1] += field[i, j, 1]
                cnt[ii, jj] += 1
        for i in range(nnr):
            for j in range(nnc):
                if cnt[i, j] > 0:
                    down[i, j, 0] /= cnt[i, j]
                    down[i, j, 1] /= cnt[i, j]
    return np.asarray(down)


def warp_3d(floating[:, :, :] volume, floating[:, :, :, :] d1,
            double[:, :] affine_idx_in=None,
            double[:, :] affine_idx_out=None,
            double[:, :] affine_disp=None,
            int[:] out_shape=None):
    r"""Warps a 3D volume using trilinear interpolation

    Deforms the input volume under the given transformation. The warped volume
    is computed using tri-linear interpolation and is given by:

    (1) warped[i] = volume[ C * d1[A*i] + B*i ]

    where A = affine_idx_in, B = affine_idx_out, C = affine_disp and i denotes
    the discrete coordinates of a voxel in the sampling grid of
    shape = out_shape.

    Parameters
    ----------
    volume : array, shape (S, R, C)
        the input volume to be transformed
    d1 : array, shape (S', R', C', 3)
        the displacement field driving the transformation
    affine_idx_in : array, shape (4, 4)
        the matrix A in eq. (1) above
    affine_idx_out : array, shape (4, 4)
        the matrix B in eq. (1) above
    affine_disp : array, shape (4, 4)
        the matrix C in eq. (1) above
    out_shape : array, shape (3,)
        the number of slices, rows and columns of the sampling grid

    Returns
    -------
    warped : array, shape = out_shape
        the transformed volume

    Notes
    -----
    To illustrate the use of this function, consider a displacement field d1
    with grid-to-space transformation R, a volume with grid-to-space
    transformation T and let's say we want to sample the warped volume on a
    grid with grid-to-space transformation S (sampling grid). For each voxel
    in the sampling grid with discrete coordinates i, the warped volume is
    given by:

    (2) warped[i] = volume[Tinv * ( d1[Rinv * S * i] + S * i ) ]

    where Tinv = T^{-1} and Rinv = R^{-1}. By identifying A = Rinv * S,
    B = Tinv * S, C = Tinv we can use this function to efficiently warp the
    input image.
    """
    cdef:
        cnp.npy_intp nslices = volume.shape[0]
        cnp.npy_intp nrows = volume.shape[1]
        cnp.npy_intp ncols = volume.shape[2]
        cnp.npy_intp nsVol = volume.shape[0]
        cnp.npy_intp nrVol = volume.shape[1]
        cnp.npy_intp ncVol = volume.shape[2]
        cnp.npy_intp i, j, k
        int inside
        double dkk, dii, djj, dk, di, dj

    if not is_valid_affine(affine_idx_in, 3):
        raise ValueError("Invalid inner index multiplication matrix")
    if not is_valid_affine(affine_idx_out, 3):
        raise ValueError("Invalid outer index multiplication matrix")
    if not is_valid_affine(affine_disp, 3):
        raise ValueError("Invalid displacement multiplication matrix")

    if out_shape is not None:
        nslices = out_shape[0]
        nrows = out_shape[1]
        ncols = out_shape[2]
    elif d1 is not None:
        nslices = d1.shape[0]
        nrows = d1.shape[1]
        ncols = d1.shape[2]

    cdef floating[:, :, :] warped = np.zeros(shape=(nslices, nrows, ncols),
                                             dtype=np.asarray(volume).dtype)
    cdef floating[:] tmp = np.zeros(shape=(3,), dtype = np.asarray(d1).dtype)

    with nogil:

        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    if affine_idx_in is None:
                        dkk = d1[k, i, j, 0]
                        dii = d1[k, i, j, 1]
                        djj = d1[k, i, j, 2]
                    else:
                        dk = _apply_affine_3d_x0(
                            k, i, j, 1, affine_idx_in)
                        di = _apply_affine_3d_x1(
                            k, i, j, 1, affine_idx_in)
                        dj = _apply_affine_3d_x2(
                            k, i, j, 1, affine_idx_in)
                        inside = _interpolate_vector_3d[floating](d1, dk, di,
                                                                  dj, &tmp[0])
                        dkk = tmp[0]
                        dii = tmp[1]
                        djj = tmp[2]

                    if affine_disp is not None:
                        dk = _apply_affine_3d_x0(
                            dkk, dii, djj, 0, affine_disp)
                        di = _apply_affine_3d_x1(
                            dkk, dii, djj, 0, affine_disp)
                        dj = _apply_affine_3d_x2(
                            dkk, dii, djj, 0, affine_disp)
                    else:
                        dk = dkk
                        di = dii
                        dj = djj

                    if affine_idx_out is not None:
                        dkk = dk + _apply_affine_3d_x0(k, i, j, 1,
                                                       affine_idx_out)
                        dii = di + _apply_affine_3d_x1(k, i, j, 1,
                                                       affine_idx_out)
                        djj = dj + _apply_affine_3d_x2(k, i, j, 1,
                                                       affine_idx_out)
                    else:
                        dkk = dk + k
                        dii = di + i
                        djj = dj + j

                    inside = _interpolate_scalar_3d[floating](volume, dkk,
                                                              dii, djj,
                                                              &warped[k,i,j])
    return np.asarray(warped)


def transform_3d_affine(floating[:, :, :] volume, int[:] ref_shape,
                        double[:, :] affine):
    r"""Transforms a 3D volume by an affine transform with trilinear interp.

    Deforms the input volume under the given affine transformation using
    tri-linear interpolation. The shape of the resulting transformation
    is given by ref_shape. If the affine matrix is None, it is taken as the
    identity.

    Parameters
    ----------
    volume : array, shape (S, R, C)
        the input volume to be transformed
    ref_shape : array, shape (3,)
        the shape of the resulting volume
    affine : array, shape (4, 4)
        the affine transform to be applied

    Returns
    -------
    out : array, shape (S', R', C')
        the transformed volume

    Notes
    -----
    The reason it is necessary to provide the intended shape of the resulting
    volume is because the affine transformation is defined on all R^{3}
    but we must sample a finite lattice. Also the resulting shape may not be
    necessarily equal to the input shape, unless we are interested on
    endomorphisms only and not general diffeomorphisms.
    """
    cdef:
        cnp.npy_intp nslices = ref_shape[0]
        cnp.npy_intp nrows = ref_shape[1]
        cnp.npy_intp ncols = ref_shape[2]
        cnp.npy_intp nsVol = volume.shape[0]
        cnp.npy_intp nrVol = volume.shape[1]
        cnp.npy_intp ncVol = volume.shape[2]
        cnp.npy_intp i, j, k, ii, jj, kk
        int inside
        double dkk, dii, djj, tmp0, tmp1
        double alpha, beta, gamma, calpha, cbeta, cgamma
        floating[:, :, :] out = np.zeros(shape=(nslices, nrows, ncols),
                                         dtype=np.asarray(volume).dtype)

    if not is_valid_affine(affine, 3):
        raise ValueError("Invalid affine transform matrix")

    with nogil:

        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    if affine is not None:
                        dkk = _apply_affine_3d_x0(k, i, j, 1, affine)
                        dii = _apply_affine_3d_x1(k, i, j, 1, affine)
                        djj = _apply_affine_3d_x2(k, i, j, 1, affine)
                    else:
                        dkk = k
                        dii = i
                        djj = j
                    inside = _interpolate_scalar_3d[floating](volume, dkk,
                        dii, djj, &out[k,i,j])
    return np.asarray(out)


def warp_3d_nn(number[:, :, :] volume, floating[:, :, :, :] d1,
               double[:, :] affine_idx_in=None,
               double[:, :] affine_idx_out=None,
               double[:, :] affine_disp=None,
               int[:] out_shape=None):
    r"""Warps a 3D volume using using nearest-neighbor interpolation

    Deforms the input volume under the given transformation. The warped volume
    is computed using nearest-neighbor interpolation and is given by:

    (1) warped[i] = volume[ C * d1[A*i] + B*i ]

    where A = affine_idx_in, B = affine_idx_out, C = affine_disp and i denotes
    the discrete coordinates of a voxel in the sampling grid of
    shape = out_shape.

    Parameters
    ----------
    volume : array, shape (S, R, C)
        the input volume to be transformed
    d1 : array, shape (S', R', C', 3)
        the displacement field driving the transformation
    affine_idx_in : array, shape (4, 4)
        the matrix A in eq. (1) above
    affine_idx_out : array, shape (4, 4)
        the matrix B in eq. (1) above
    affine_disp : array, shape (4, 4)
        the matrix C in eq. (1) above
    out_shape : array, shape (3,)
        the number of slices, rows and columns of the sampling grid

    Returns
    -------
    warped : array, shape = out_shape
        the transformed volume

    Notes
    -----
    To illustrate the use of this function, consider a displacement field d1
    with grid-to-space transformation R, a volume with grid-to-space
    transformation T and let's say we want to sample the warped volume on a
    grid with grid-to-space transformation S (sampling grid). For each voxel
    in the sampling grid with discrete coordinates i, the warped volume is
    given by:

    (2) warped[i] = volume[Tinv * ( d1[Rinv * S * i] + S * i ) ]

    where Tinv = T^{-1} and Rinv = R^{-1}. By identifying A = Rinv * S,
    B = Tinv * S, C = Tinv we can use this function to efficiently warp the
    input image.
    """
    cdef:
        cnp.npy_intp nslices = volume.shape[0]
        cnp.npy_intp nrows = volume.shape[1]
        cnp.npy_intp ncols = volume.shape[2]
        cnp.npy_intp nsVol = volume.shape[0]
        cnp.npy_intp nrVol = volume.shape[1]
        cnp.npy_intp ncVol = volume.shape[2]
        cnp.npy_intp i, j, k
        int inside
        double dkk, dii, djj, dk, di, dj

    if not is_valid_affine(affine_idx_in, 3):
        raise ValueError("Invalid inner index multiplication matrix")
    if not is_valid_affine(affine_idx_out, 3):
        raise ValueError("Invalid outer index multiplication matrix")
    if not is_valid_affine(affine_disp, 3):
        raise ValueError("Invalid displacement multiplication matrix")

    if out_shape is not None:
        nslices = out_shape[0]
        nrows = out_shape[1]
        ncols = out_shape[2]
    elif d1 is not None:
        nslices = d1.shape[0]
        nrows = d1.shape[1]
        ncols = d1.shape[2]

    cdef number[:, :, :] warped = np.zeros(shape=(nslices, nrows, ncols),
                                           dtype=np.asarray(volume).dtype)
    cdef floating[:] tmp = np.zeros(shape=(3,), dtype = np.asarray(d1).dtype)

    with nogil:

        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    if affine_idx_in is None:
                        dkk = d1[k, i, j, 0]
                        dii = d1[k, i, j, 1]
                        djj = d1[k, i, j, 2]
                    else:
                        dk = _apply_affine_3d_x0(
                            k, i, j, 1, affine_idx_in)
                        di = _apply_affine_3d_x1(
                            k, i, j, 1, affine_idx_in)
                        dj = _apply_affine_3d_x2(
                            k, i, j, 1, affine_idx_in)
                        inside = _interpolate_vector_3d[floating](d1, dk, di,
                                                                  dj, &tmp[0])
                        dkk = tmp[0]
                        dii = tmp[1]
                        djj = tmp[2]

                    if affine_disp is not None:
                        dk = _apply_affine_3d_x0(
                            dkk, dii, djj, 0, affine_disp)
                        di = _apply_affine_3d_x1(
                            dkk, dii, djj, 0, affine_disp)
                        dj = _apply_affine_3d_x2(
                            dkk, dii, djj, 0, affine_disp)
                    else:
                        dk = dkk
                        di = dii
                        dj = djj

                    if affine_idx_out is not None:
                        dkk = dk + _apply_affine_3d_x0(k, i, j, 1,
                                                       affine_idx_out)
                        dii = di + _apply_affine_3d_x1(k, i, j, 1,
                                                       affine_idx_out)
                        djj = dj + _apply_affine_3d_x2(k, i, j, 1,
                                                       affine_idx_out)
                    else:
                        dkk = dk + k
                        dii = di + i
                        djj = dj + j

                    inside = _interpolate_scalar_nn_3d[number](volume, dkk, dii, djj,
                                                       &warped[k,i,j])
    return np.asarray(warped)


def transform_3d_affine_nn(number[:, :, :] volume, int[:] ref_shape,
                           double[:, :] affine=None):
    r"""Transforms a 3D volume by an affine transform with NN interpolation

    Deforms the input volume under the given affine transformation using
    nearest neighbor interpolation. The shape of the resulting volume
    is given by ref_shape. If the affine matrix is None, it is taken as the
    identity.

    Parameters
    ----------
    volume : array, shape (S, R, C)
        the input volume to be transformed
    ref_shape : array, shape (3,)
        the shape of the resulting volume
    affine : array, shape (4, 4)
        the affine transform to be applied

    Returns
    -------
    out : array, shape (S', R', C')
        the transformed volume

    Notes
    -----
    The reason it is necessary to provide the intended shape of the resulting
    volume is because the affine transformation is defined on all R^{3}
    but we must sample a finite lattice. Also the resulting shape may not be
    necessarily equal to the input shape, unless we are interested on
    endomorphisms only and not general diffeomorphisms.
    """
    cdef:
        cnp.npy_intp nslices = ref_shape[0]
        cnp.npy_intp nrows = ref_shape[1]
        cnp.npy_intp ncols = ref_shape[2]
        cnp.npy_intp nsVol = volume.shape[0]
        cnp.npy_intp nrVol = volume.shape[1]
        cnp.npy_intp ncVol = volume.shape[2]
        double dkk, dii, djj, tmp0, tmp1
        double alpha, beta, gamma, calpha, cbeta, cgamma
        cnp.npy_intp k, i, j, kk, ii, jj
        number[:, :, :] out = np.zeros((nslices, nrows, ncols),
                                        dtype=np.asarray(volume).dtype)

    if not is_valid_affine(affine, 3):
        raise ValueError("Invalid affine transform matrix")

    with nogil:

        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    if affine is not None:
                        dkk = _apply_affine_3d_x0(k, i, j, 1, affine)
                        dii = _apply_affine_3d_x1(k, i, j, 1, affine)
                        djj = _apply_affine_3d_x2(k, i, j, 1, affine)
                    else:
                        dkk = k
                        dii = i
                        djj = j
                    _interpolate_scalar_nn_3d[number](volume, dkk, dii, djj,
                                                      &out[k,i,j])
    return np.asarray(out)


def warp_2d(floating[:, :] image, floating[:, :, :] d1,
            double[:, :] affine_idx_in=None,
            double[:, :] affine_idx_out=None,
            double[:, :] affine_disp=None,
            int[:] out_shape=None):
    r"""Warps a 2D image using bilinear interpolation

    Deforms the input image under the given transformation. The warped image
    is computed using bi-linear interpolation and is given by:

    (1) warped[i] = image[ C * d1[A*i] + B*i ]

    where A = affine_idx_in, B = affine_idx_out, C = affine_disp and i denotes
    the discrete coordinates of a voxel in the sampling grid of
    shape = out_shape.

    Parameters
    ----------
    image : array, shape (R, C)
        the input image to be transformed
    d1 : array, shape (R', C', 2)
        the displacement field driving the transformation
    affine_idx_in : array, shape (3, 3)
        the matrix A in eq. (1) above
    affine_idx_out : array, shape (3, 3)
        the matrix B in eq. (1) above
    affine_disp : array, shape (3, 3)
        the matrix C in eq. (1) above
    out_shape : array, shape (2,)
        the number of rows and columns of the sampling grid

    Returns
    -------
    warped : array, shape = out_shape
        the transformed image

    Notes
    -----
    To illustrate the use of this function, consider a displacement field d1
    with grid-to-space transformation R, an image with grid-to-space
    transformation T and let's say we want to sample the warped image on a
    grid with grid-to-space transformation S (sampling grid). For each voxel
    in the sampling grid with discrete coordinates i, the warped image is
    given by:

    (2) warped[i] = image[Tinv * ( d1[Rinv * S * i] + S * i ) ]

    where Tinv = T^{-1} and Rinv = R^{-1}. By identifying A = Rinv * S,
    B = Tinv * S, C = Tinv we can use this function to efficiently warp the
    input image.
    """
    cdef:
        cnp.npy_intp nrows = image.shape[0]
        cnp.npy_intp ncols = image.shape[1]
        cnp.npy_intp nrVol = image.shape[0]
        cnp.npy_intp ncVol = image.shape[1]
        cnp.npy_intp i, j, ii, jj
        double di, dj, dii, djj

    if not is_valid_affine(affine_idx_in, 2):
        raise ValueError("Invalid inner index multiplication matrix")
    if not is_valid_affine(affine_idx_out, 2):
        raise ValueError("Invalid outer index multiplication matrix")
    if not is_valid_affine(affine_disp, 2):
        raise ValueError("Invalid displacement multiplication matrix")

    if out_shape is not None:
        nrows = out_shape[0]
        ncols = out_shape[1]
    elif d1 is not None:
        nrows = d1.shape[0]
        ncols = d1.shape[1]
    cdef floating[:, :] warped = np.zeros(shape=(nrows, ncols),
                                          dtype=np.asarray(image).dtype)
    cdef floating[:] tmp = np.zeros(shape=(2,), dtype=np.asarray(d1).dtype)

    with nogil:

        for i in range(nrows):
            for j in range(ncols):
                # Apply inner index pre-multiplication
                if affine_idx_in is None:
                    dii = d1[i, j, 0]
                    djj = d1[i, j, 1]
                else:
                    di = _apply_affine_2d_x0(
                        i, j, 1, affine_idx_in)
                    dj = _apply_affine_2d_x1(
                        i, j, 1, affine_idx_in)
                    _interpolate_vector_2d[floating](d1, di, dj, &tmp[0])
                    dii = tmp[0]
                    djj = tmp[1]

                # Apply displacement multiplication
                if affine_disp is not None:
                    di = _apply_affine_2d_x0(
                        dii, djj, 0, affine_disp)
                    dj = _apply_affine_2d_x1(
                        dii, djj, 0, affine_disp)
                else:
                    di = dii
                    dj = djj

                # Apply outer index multiplication and add the displacements
                if affine_idx_out is not None:
                    dii = di + _apply_affine_2d_x0(i, j, 1, affine_idx_out)
                    djj = dj + _apply_affine_2d_x1(i, j, 1, affine_idx_out)
                else:
                    dii = di + i
                    djj = dj + j

                # Interpolate the input image at the resulting location
                _interpolate_scalar_2d[floating](image, dii, djj,
                                                 &warped[i, j])
    return np.asarray(warped)


def transform_2d_affine(floating[:, :] image, int[:] ref_shape,
                        double[:, :] affine=None):
    r"""Transforms a 2D image by an affine transform with bilinear interp.

    Deforms the input image under the given affine transformation using
    tri-linear interpolation. The shape of the resulting image
    is given by ref_shape. If the affine matrix is None, it is taken as the
    identity.

    Parameters
    ----------
    image : array, shape (R, C)
        the input image to be transformed
    ref_shape : array, shape (2,)
        the shape of the resulting image
    affine : array, shape (3, 3)
        the affine transform to be applied

    Returns
    -------
    out : array, shape (R', C')
        the transformed image

    Notes
    -----
    The reason it is necessary to provide the intended shape of the resulting
    image is because the affine transformation is defined on all R^{2}
    but we must sample a finite lattice. Also the resulting shape may not be
    necessarily equal to the input shape, unless we are interested on
    endomorphisms only and not general diffeomorphisms.
    """
    cdef:
        cnp.npy_intp nrows = ref_shape[0]
        cnp.npy_intp ncols = ref_shape[1]
        cnp.npy_intp nrVol = image.shape[0]
        cnp.npy_intp ncVol = image.shape[1]
        cnp.npy_intp i, j, ii, jj
        double dii, djj, tmp0
        double alpha, beta, calpha, cbeta
        floating[:, :] out = np.zeros(shape=(nrows, ncols),
                                      dtype=np.asarray(image).dtype)

    if not is_valid_affine(affine, 2):
        raise ValueError("Invalid affine transform matrix")

    with nogil:

        for i in range(nrows):
            for j in range(ncols):
                if affine is not None:
                    dii = _apply_affine_2d_x0(i, j, 1, affine)
                    djj = _apply_affine_2d_x1(i, j, 1, affine)
                else:
                    dii = i
                    djj = j
                _interpolate_scalar_2d[floating](image, dii, djj,
                                                 &out[i, j])
    return np.asarray(out)


def warp_2d_nn(number[:, :] image, floating[:, :, :] d1,
               double[:, :] affine_idx_in=None,
               double[:, :] affine_idx_out=None,
               double[:, :] affine_disp=None,
               int[:] out_shape=None):
    r"""Warps a 2D image using nearest neighbor interpolation

    Deforms the input image under the given transformation. The warped image
    is computed using nearest-neighbor interpolation and is given by:

    (1) warped[i] = image[ C * d1[A*i] + B*i ]

    where A = affine_idx_in, B = affine_idx_out, C = affine_disp and i denotes
    the discrete coordinates of a voxel in the sampling grid of
    shape = out_shape.

    Parameters
    ----------
    image : array, shape (R, C)
        the input image to be transformed
    d1 : array, shape (R', C', 2)
        the displacement field driving the transformation
    affine_idx_in : array, shape (3, 3)
        the matrix A in eq. (1) above
    affine_idx_out : array, shape (3, 3)
        the matrix B in eq. (1) above
    affine_disp : array, shape (3, 3)
        the matrix C in eq. (1) above
    out_shape : array, shape (2,)
        the number of rows and columns of the sampling grid

    Returns
    -------
    warped : array, shape = out_shape
        the transformed image

    Notes
    -----
    To illustrate the use of this function, consider a displacement field d1
    with grid-to-space transformation R, an image with grid-to-space
    transformation T and let's say we want to sample the warped image on a
    grid with grid-to-space transformation S (sampling grid). For each voxel
    in the sampling grid with discrete coordinates i, the warped image is
    given by:

    (2) warped[i] = image[Tinv * ( d1[Rinv * S * i] + S * i ) ]

    where Tinv = T^{-1} and Rinv = R^{-1}. By identifying A = Rinv * S,
    B = Tinv * S, C = Tinv we can use this function to efficiently warp the
    input image.
    """
    cdef:
        cnp.npy_intp nrows = image.shape[0]
        cnp.npy_intp ncols = image.shape[1]
        cnp.npy_intp nrVol = image.shape[0]
        cnp.npy_intp ncVol = image.shape[1]
        cnp.npy_intp i, j, ii, jj
        double di, dj, dii, djj

    if not is_valid_affine(affine_idx_in, 2):
        raise ValueError("Invalid inner index multiplication matrix")
    if not is_valid_affine(affine_idx_out, 2):
        raise ValueError("Invalid outer index multiplication matrix")
    if not is_valid_affine(affine_disp, 2):
        raise ValueError("Invalid displacement multiplication matrix")

    if out_shape is not None:
        nrows = out_shape[0]
        ncols = out_shape[1]
    elif d1 is not None:
        nrows = d1.shape[0]
        ncols = d1.shape[1]
    cdef number[:, :] warped = np.zeros(shape=(nrows, ncols),
                                        dtype=np.asarray(image).dtype)
    cdef floating[:] tmp = np.zeros(shape=(2,), dtype=np.asarray(d1).dtype)

    with nogil:

        for i in range(nrows):
            for j in range(ncols):
                # Apply inner index pre-multiplication
                if affine_idx_in is None:
                    dii = d1[i, j, 0]
                    djj = d1[i, j, 1]
                else:
                    di = _apply_affine_2d_x0(
                        i, j, 1, affine_idx_in)
                    dj = _apply_affine_2d_x1(
                        i, j, 1, affine_idx_in)
                    _interpolate_vector_2d[floating](d1, di, dj, &tmp[0])
                    dii = tmp[0]
                    djj = tmp[1]

                # Apply displacement multiplication
                if affine_disp is not None:
                    di = _apply_affine_2d_x0(
                        dii, djj, 0, affine_disp)
                    dj = _apply_affine_2d_x1(
                        dii, djj, 0, affine_disp)
                else:
                    di = dii
                    dj = djj

                # Apply outer index multiplication and add the displacements
                if affine_idx_out is not None:
                    dii = di + _apply_affine_2d_x0(i, j, 1, affine_idx_out)
                    djj = dj + _apply_affine_2d_x1(i, j, 1, affine_idx_out)
                else:
                    dii = di + i
                    djj = dj + j

                # Interpolate the input image at the resulting location
                _interpolate_scalar_nn_2d[number](image, dii, djj,
                                                  &warped[i, j])
    return np.asarray(warped)


def transform_2d_affine_nn(number[:, :] image, int[:] ref_shape,
                           double[:, :] affine=None):
    r"""Transforms a 2D image by an affine transform with NN interpolation

    Deforms the input image under the given affine transformation using
    nearest neighbor interpolation. The shape of the resulting image
    is given by ref_shape. If the affine matrix is None, it is taken as the
    identity.

    Parameters
    ----------
    image : array, shape (R, C)
        the input image to be transformed
    ref_shape : array, shape (2,)
        the shape of the resulting image
    affine : array, shape (3, 3)
        the affine transform to be applied

    Returns
    -------
    out : array, shape (R', C')
        the transformed image

    Notes
    -----
    The reason it is necessary to provide the intended shape of the resulting
    image is because the affine transformation is defined on all R^{2}
    but we must sample a finite lattice. Also the resulting shape may not be
    necessarily equal to the input shape, unless we are interested on
    endomorphisms only and not general diffeomorphisms.
    """
    cdef:
        cnp.npy_intp nrows = ref_shape[0]
        cnp.npy_intp ncols = ref_shape[1]
        cnp.npy_intp nrVol = image.shape[0]
        cnp.npy_intp ncVol = image.shape[1]
        double dii, djj, tmp0
        double alpha, beta, calpha, cbeta
        cnp.npy_intp i, j, ii, jj
        number[:, :] out = np.zeros((nrows, ncols),
                                    dtype=np.asarray(image).dtype)

    if not is_valid_affine(affine, 2):
        raise ValueError("Invalid affine transform matrix")

    with nogil:

        for i in range(nrows):
            for j in range(ncols):
                if affine is not None:
                    dii = _apply_affine_2d_x0(i, j, 1, affine)
                    djj = _apply_affine_2d_x1(i, j, 1, affine)
                else:
                    dii = i
                    djj = j
                _interpolate_scalar_nn_2d[number](image, dii, djj,
                                                  &out[i, j])
    return np.asarray(out)


def resample_displacement_field_3d(floating[:, :, :, :] field,
                                   double[:] factors, int[:] out_shape):
    r"""Resamples a 3D vector field to a custom target shape

    Resamples the given 3D displacement field on a grid of the requested shape,
    using the given scale factors. More precisely, the resulting displacement
    field at each grid cell i is given by

    D[i] = field[Diag(factors) * i]

    Parameters
    ----------
    factors : array, shape (3,)
        the scaling factors mapping (integer) grid coordinates in the resampled
        grid to (floating point) grid coordinates in the original grid
    out_shape : array, shape (3,)
        the desired shape of the resulting grid

    Returns
    -------
    expanded : array, shape = out_shape + (3, )
        the resampled displacement field
    """
    ftype = np.asarray(field).dtype
    cdef:
        cnp.npy_intp tslices = out_shape[0]
        cnp.npy_intp trows = out_shape[1]
        cnp.npy_intp tcols = out_shape[2]
        cnp.npy_intp k, i, j
        int inside
        double dkk, dii, djj
        floating[:, :, :, :] expanded = np.zeros((tslices, trows, tcols, 3),
                                                 dtype=ftype)

    for k in range(tslices):
        for i in range(trows):
            for j in range(tcols):
                dkk = <double> k * factors[0]
                dii = <double> i * factors[1]
                djj = <double> j * factors[2]
                _interpolate_vector_3d[floating](field, dkk, dii, djj,
                                                 &expanded[k, i, j, 0])
    return np.asarray(expanded)


def resample_displacement_field_2d(floating[:, :, :] field, double[:] factors,
                                   int[:] out_shape):
    r"""Resamples a 2D vector field to a custom target shape

    Resamples the given 2D displacement field on a grid of the requested shape,
    using the given scale factors. More precisely, the resulting displacement
    field at each grid cell i is given by

    D[i] = field[Diag(factors) * i]

    Parameters
    ----------
    factors : array, shape (2,)
        the scaling factors mapping (integer) grid coordinates in the resampled
        grid to (floating point) grid coordinates in the original grid
    out_shape : array, shape (2,)
        the desired shape of the resulting grid

    Returns
    -------
    expanded : array, shape = out_shape + (2, )
        the resampled displacement field
    """
    ftype = np.asarray(field).dtype
    cdef:
        cnp.npy_intp trows = out_shape[0]
        cnp.npy_intp tcols = out_shape[1]
        cnp.npy_intp i, j
        int inside
        double dii, djj
        floating[:, :, :] expanded = np.zeros((trows, tcols, 2), dtype=ftype)

    for i in range(trows):
        for j in range(tcols):
            dii = i*factors[0]
            djj = j*factors[1]
            inside = _interpolate_vector_2d[floating](field, dii, djj,
                                                      &expanded[i, j, 0])
    return np.asarray(expanded)


def create_random_displacement_2d(int[:] from_shape,
                                  double[:, :] from_grid2world,
                                  int[:] to_shape,
                                  double[:, :] to_grid2world):
    r"""Creates a random 2D displacement 'exactly' mapping points of two grids

    Creates a random 2D displacement field mapping points of an input discrete
    domain (with dimensions given by from_shape) to points of an output
    discrete domain (with shape given by to_shape). The affine matrices
    bringing discrete coordinates to physical space are given by from_grid2world
    (for the displacement field discretization) and to_grid2world (for the target
    discretization). Since this function is intended to be used for testing,
    voxels in the input domain will never be assigned to boundary voxels on the
    output domain.

    Parameters
    ----------
    from_shape : array, shape (2,)
        the grid shape where the displacement field will be defined on.
    from_grid2world : array, shape (3,3)
        the grid-to-space transformation of the displacement field
    to_shape : array, shape (2,)
        the grid shape where the deformation field will map the input grid to.
    to_grid2world : array, shape (3,3)
        the grid-to-space transformation of the mapped grid

    Returns
    -------
    output : array, shape = from_shape
        the random displacement field in the physical domain
    int_field : array, shape = from_shape
        the assignment of each point in the input grid to the target grid
    """
    cdef:
        cnp.npy_intp i, j, ri, rj
        double di, dj, dii, djj
        int[:, :, :] int_field = np.empty(tuple(from_shape) + (2,),
                                            dtype=np.int32)
        double[:, :, :] output = np.zeros(tuple(from_shape) + (2,),
                                          dtype=np.float64)
        cnp.npy_intp dom_size = from_shape[0]*from_shape[1]

    if not is_valid_affine(from_grid2world, 2):
        raise ValueError("Invalid 'from' affine transform matrix")
    if not is_valid_affine(to_grid2world, 2):
        raise ValueError("Invalid 'to' affine transform matrix")

    # compute the actual displacement field in the physical space
    for i in range(from_shape[0]):
        for j in range(from_shape[1]):
            # randomly choose where each input grid point will be mapped to in
            # the target grid
            ri = np.random.randint(1, to_shape[0]-1)
            rj = np.random.randint(1, to_shape[1]-1)
            int_field[i, j, 0] = ri
            int_field[i, j, 1] = rj

            # convert the input point to physical coordinates
            if from_grid2world is not None:
                di = _apply_affine_2d_x0(i, j, 1, from_grid2world)
                dj = _apply_affine_2d_x1(i, j, 1, from_grid2world)
            else:
                di = i
                dj = j

            # convert the output point to physical coordinates
            if to_grid2world is not None:
                dii = _apply_affine_2d_x0(ri, rj, 1, to_grid2world)
                djj = _apply_affine_2d_x1(ri, rj, 1, to_grid2world)
            else:
                dii = ri
                djj = rj

            # the displacement vector at (i,j) must be the target point minus
            # the original point, both in physical space

            output[i, j, 0] = dii - di
            output[i, j, 1] = djj - dj

    return np.asarray(output), np.asarray(int_field)


def create_random_displacement_3d(int[:] from_shape, double[:, :] from_grid2world,
                                  int[:] to_shape, double[:, :] to_grid2world):
    r"""Creates a random 3D displacement 'exactly' mapping points of two grids

    Creates a random 3D displacement field mapping points of an input discrete
    domain (with dimensions given by from_shape) to points of an output
    discrete domain (with shape given by to_shape). The affine matrices
    bringing discrete coordinates to physical space are given by from_grid2world
    (for the displacement field discretization) and to_grid2world (for the target
    discretization). Since this function is intended to be used for testing,
    voxels in the input domain will never be assigned to boundary voxels on the
    output domain.

    Parameters
    ----------
    from_shape : array, shape (3,)
        the grid shape where the displacement field will be defined on.
    from_grid2world : array, shape (4,4)
        the grid-to-space transformation of the displacement field
    to_shape : array, shape (3,)
        the grid shape where the deformation field will map the input grid to.
    to_grid2world : array, shape (4,4)
        the grid-to-space transformation of the mapped grid

    Returns
    -------
    output : array, shape = from_shape
        the random displacement field in the physical domain
    int_field : array, shape = from_shape
        the assignment of each point in the input grid to the target grid
    """
    cdef:
        cnp.npy_intp i, j, k, ri, rj, rk
        double di, dj, dk, dii, djj, dkk
        int[:, :, :, :] int_field = np.empty(tuple(from_shape) + (3,),
                                               dtype=np.int32)
        double[:, :, :, :] output = np.zeros(tuple(from_shape) + (3,),
                                             dtype=np.float64)
        cnp.npy_intp dom_size = from_shape[0]*from_shape[1]*from_shape[2]

    if not is_valid_affine(from_grid2world, 3):
        raise ValueError("Invalid 'from' affine transform matrix")
    if not is_valid_affine(to_grid2world, 3):
        raise ValueError("Invalid 'to' affine transform matrix")

    # compute the actual displacement field in the physical space
    for k in range(from_shape[0]):
        for i in range(from_shape[1]):
            for j in range(from_shape[2]):
                # randomly choose the location of each point on the target grid
                rk = np.random.randint(1, to_shape[0]-1)
                ri = np.random.randint(1, to_shape[1]-1)
                rj = np.random.randint(1, to_shape[2]-1)
                int_field[k, i, j, 0] = rk
                int_field[k, i, j, 1] = ri
                int_field[k, i, j, 2] = rj

                # convert the input point to physical coordinates
                if from_grid2world is not None:
                    dk = _apply_affine_3d_x0(k, i, j, 1, from_grid2world)
                    di = _apply_affine_3d_x1(k, i, j, 1, from_grid2world)
                    dj = _apply_affine_3d_x2(k, i, j, 1, from_grid2world)
                else:
                    dk = k
                    di = i
                    dj = j

                # convert the output point to physical coordinates
                if to_grid2world is not None:
                    dkk = _apply_affine_3d_x0(rk, ri, rj, 1, to_grid2world)
                    dii = _apply_affine_3d_x1(rk, ri, rj, 1, to_grid2world)
                    djj = _apply_affine_3d_x2(rk, ri, rj, 1, to_grid2world)
                else:
                    dkk = rk
                    dii = ri
                    djj = rj

                # the displacement vector at (i,j) must be the target point minus
                # the original point, both in physical space

                output[k, i, j, 0] = dkk - dk
                output[k, i, j, 1] = dii - di
                output[k, i, j, 2] = djj - dj

    return np.asarray(output), np.asarray(int_field)


def create_harmonic_fields_2d(cnp.npy_intp nrows, cnp.npy_intp ncols,
                              double b, double m):
    r"""Creates an invertible 2D displacement field

    Creates the invertible displacement fields used in Chen et al. eqs.
    9 and 10 [1]

    Parameters
    ----------
    nrows : int
        number of rows in the resulting harmonic field
    ncols : int
        number of columns in the resulting harmonic field
    b, m : float
        parameters of the harmonic field (as in [1]). To understand the effect
        of these parameters, please consider plotting a deformed image
        (a circle or a grid) under the deformation field, or see examples
        in [1]

    Returns
    -------
    d : array, shape (nrows, ncols, 2)
        the harmonic displacement field
    inv : array, shape (nrows, ncols, 2)
        the analytical inverse of the harmonic displacement field

    [1] Chen, M., Lu, W., Chen, Q., Ruchala, K. J., & Olivera, G. H. (2008).
        A simple fixed-point approach to invert a deformation field.
        Medical Physics, 35(1), 81. doi:10.1118/1.2816107
    """
    cdef:
        cnp.npy_intp mid_row = nrows/2
        cnp.npy_intp mid_col = ncols/2
        cnp.npy_intp i, j, ii, jj
        double theta
        double[:, :, :] d = np.zeros((nrows, ncols, 2), dtype=np.float64)
        double[:, :, :] inv = np.zeros((nrows, ncols, 2), dtype=np.float64)
    for i in range(nrows):
        for j in range(ncols):
            ii = i - mid_row
            jj = j - mid_col
            theta = atan2(ii, jj)
            d[i, j, 0] = ii * (1.0 / (1 + b * cos(m * theta)) - 1.0)
            d[i, j, 1] = jj * (1.0 / (1 + b * cos(m * theta)) - 1.0)
            inv[i, j, 0] = b * cos(m * theta) * ii
            inv[i, j, 1] = b * cos(m * theta) * jj

    return np.asarray(d), np.asarray(inv)


def create_harmonic_fields_3d(int nslices, cnp.npy_intp nrows,
                              cnp.npy_intp ncols, double b, double m):
    r"""Creates an invertible 3D displacement field

    Creates the invertible displacement fields used in Chen et al. eqs.
    9 and 10 [1] computing the angle theta along z-slides.

    Parameters
    ----------
    nslices : int
        number of slices in the resulting harmonic field
    nrows : int
        number of rows in the resulting harmonic field
    ncols : int
        number of columns in the resulting harmonic field
    b, f : float
        parameters of the harmonic field (as in [1]). To understand the effect
        of these parameters, please consider plotting a deformed image
        (e.g. a circle or a grid) under the deformation field, or see examples
        in [1]

    Returns
    -------
    d : array, shape (nslices, nrows, ncols, 3)
        the harmonic displacement field
    inv : array, shape (nslices, nrows, ncols, 3)
        the analytical inverse of the harmonic displacement field

    [1] Chen, M., Lu, W., Chen, Q., Ruchala, K. J., & Olivera, G. H. (2008).
        A simple fixed-point approach to invert a deformation field.
        Medical Physics, 35(1), 81. doi:10.1118/1.2816107
    """
    cdef:
        cnp.npy_intp mid_slice = nslices / 2
        cnp.npy_intp mid_row = nrows / 2
        cnp.npy_intp mid_col = ncols / 2
        cnp.npy_intp i, j, k, ii, jj, kk
        double theta
        double[:, :, :, :] d = np.zeros((nslices, nrows, ncols, 3),
                                        dtype=np.float64)
        double[:, :, :, :] inv = np.zeros((nslices, nrows, ncols, 3),
                                          dtype=np.float64)
    for k in range(nslices):
        for i in range(nrows):
            for j in range(ncols):
                kk = k - mid_slice
                ii = i - mid_row
                jj = j - mid_col
                theta = atan2(ii, jj)
                d[k, i, j, 0] = kk * (1.0 / (1 + b * cos(m * theta)) - 1.0)
                d[k, i, j, 1] = ii * (1.0 / (1 + b * cos(m * theta)) - 1.0)
                d[k, i, j, 2] = jj * (1.0 / (1 + b * cos(m * theta)) - 1.0)
                inv[k, i, j, 0] = b * cos(m * theta) * kk
                inv[k, i, j, 1] = b * cos(m * theta) * ii
                inv[k, i, j, 2] = b * cos(m * theta) * jj

    return np.asarray(d), np.asarray(inv)


def create_circle(cnp.npy_intp nrows, cnp.npy_intp ncols, cnp.npy_intp radius):
    r"""
    Create a binary 2D image where pixel values are 1 iff their distance
    to the center of the image is less than or equal to radius.

    Parameters
    ----------
    nrows : int
        number of rows of the resulting image
    ncols : int
        number of columns of the resulting image
    radius : int
        the radius of the circle

    Returns
    -------
    c : array, shape (nrows, ncols)
        the binary image of the circle with the requested dimensions
    """
    cdef:
        cnp.npy_intp mid_row = nrows/2
        cnp.npy_intp mid_col = ncols/2
        cnp.npy_intp i, j, ii, jj
        double r
        double[:, :] c = np.zeros((nrows, ncols), dtype=np.float64)
    for i in range(nrows):
        for j in range(ncols):
            ii = i - mid_row
            jj = j - mid_col
            r = sqrt(ii*ii + jj*jj)
            if r <= radius:
                c[i, j] = 1
            else:
                c[i, j] = 0
    return np.asarray(c)


def create_sphere(cnp.npy_intp nslices, cnp.npy_intp nrows,
                  cnp.npy_intp ncols, cnp.npy_intp radius):
    r"""
    Create a binary 3D image where voxel values are 1 iff their distance
    to the center of the image is less than or equal to radius.

    Parameters
    ----------
    nslices : int
        number if slices of the resulting image
    nrows : int
        number of rows of the resulting image
    ncols : int
        number of columns of the resulting image
    radius : int
        the radius of the sphere

    Returns
    -------
    c : array, shape (nslices, nrows, ncols)
        the binary image of the sphere with the requested dimensions
    """
    cdef:
        cnp.npy_intp mid_slice = nslices/2
        cnp.npy_intp mid_row = nrows/2
        cnp.npy_intp mid_col = ncols/2
        cnp.npy_intp i, j, k, ii, jj, kk
        double r
        double[:, :, :] s = np.zeros((nslices, nrows, ncols), dtype=np.float64)

    for k in range(nslices):
        for i in range(nrows):
            for j in range(ncols):
                kk = k - mid_slice
                ii = i - mid_row
                jj = j - mid_col
                r = sqrt(ii*ii + jj*jj + kk*kk)
                if r <= radius:
                    s[k, i, j] = 1
                else:
                    s[k, i, j] = 0
    return np.asarray(s)


def _gradient_3d(floating[:, :, :] img, double[:, :] img_world2grid,
                 double[:] img_spacing, double[:, :] out_grid2world,
                 floating[:, :, :, :] out, int[:, :, :] inside):
    r""" Gradient of a 3D image in physical space coordinates

    Each grid cell (i, j, k) in the sampling grid (determined by
    out.shape) is mapped to its corresponding physical point (x, y, z) by
    multiplying out_grid2world (its grid-to-space transform) by (i, j, k),
    then the image is interpolated, at

    P1=(x + h, y, z), Q1=(x - h, y, z)
    P2=(x, y + h, z), Q2=(x, y - h, z)
    P3=(x, y, z + h), Q3=(x, y, z - h)

    (by mapping Pi and Qi to the grid using img_world2grid: the inverse of the
    grid-to-space transform of img). The displacement parameter h is of
    magnitude 0.5 (in physical space units), therefore the approximated partial
    derivatives are given by the difference between the image interpolated at
    Pi and Qi.

    Parameters
    ----------
    img : array, shape (S, R, C)
        the input volume whose gradient will be computed
    img_world2grid : array, shape (4, 4)
        the space-to-grid transform matrix associated to img
    img_spacing : array, shape (3,)
        the spacing between voxels (voxel size along each axis) of the input
        volume
    out_grid2world : array, shape (4, 4)
        the grid-to-space transform associated to the sampling grid
    out : array, shape (S', R', C', 3)
        the buffer in which to store the image gradient
    inside : array, shape (S', R', C')
        the buffer in which to store the flags indicating whether the sample
        point lies inside (=1) or outside (=0) the image grid
    """
    cdef:
        int nslices = out.shape[0]
        int nrows = out.shape[1]
        int ncols = out.shape[2]
        int i, j, k, in_flag
        double tmp
        double[:] x = np.empty(shape=(3,), dtype=np.float64)
        double[:] dx = np.empty(shape=(3,), dtype=np.float64)
        double[:] h = np.empty(shape=(3,), dtype=np.float64)
        double[:] q = np.empty(shape=(3,), dtype=np.float64)
    with nogil:
        h[0] = 0.5 * img_spacing[0]
        h[1] = 0.5 * img_spacing[1]
        h[2] = 0.5 * img_spacing[2]
        for k in range(nslices):
            for i in range(nrows):
                for j in range(ncols):
                    inside[k, i, j] = 1
                    # Compute coordinates of index (k, i, j) in physical space
                    x[0] = _apply_affine_3d_x0(k, i, j, 1, out_grid2world)
                    x[1] = _apply_affine_3d_x1(k, i, j, 1, out_grid2world)
                    x[2] = _apply_affine_3d_x2(k, i, j, 1, out_grid2world)
                    dx[:] = x[:]
                    for p in range(3):
                        # Compute coordinates of point dx on img's grid
                        dx[p] = x[p] - h[p]
                        q[0] = _apply_affine_3d_x0(dx[0], dx[1], dx[2], 1,
                                                   img_world2grid)
                        q[1] = _apply_affine_3d_x1(dx[0], dx[1], dx[2], 1,
                                                   img_world2grid)
                        q[2] = _apply_affine_3d_x2(dx[0], dx[1], dx[2], 1,
                                                   img_world2grid)
                        # Interpolate img at q
                        in_flag = _interpolate_scalar_3d[floating](img, q[0],
                            q[1], q[2], &out[k, i, j, p])
                        if in_flag == 0:
                            out[k, i, j, p] = 0
                            inside[k, i, j] = 0
                            continue
                        tmp = out[k, i, j, p]
                        # Compute coordinates of point dx on img's grid
                        dx[p] = x[p] + h[p]
                        q[0] = _apply_affine_3d_x0(dx[0], dx[1], dx[2], 1,
                                                   img_world2grid)
                        q[1] = _apply_affine_3d_x1(dx[0], dx[1], dx[2], 1,
                                                   img_world2grid)
                        q[2] = _apply_affine_3d_x2(dx[0], dx[1], dx[2], 1,
                                                   img_world2grid)
                        # Interpolate img at q
                        in_flag = _interpolate_scalar_3d[floating](img, q[0],
                            q[1], q[2], &out[k, i, j, p])
                        if in_flag == 0:
                            out[k, i, j, p] = 0
                            inside[k, i, j] = 0
                            continue
                        out[k, i, j, p] = (out[k, i, j, p] - tmp) / img_spacing[p]
                        dx[p] = x[p]


def _sparse_gradient_3d(floating[:, :, :] img,
                        double[:, :] img_world2grid,
                        double[:] img_spacing,
                        double[:, :] sample_points,
                        floating[:, :] out, int[:] inside):
    r""" Gradient of a 3D image evaluated at a set of points in physical space

    For each row (x_i, y_i, z_i) in sample_points, the image is interpolated at

    P1=(x_i + h, y_i, z_i), Q1=(x_i - h, y_i, z_i)
    P2=(x_i, y_i + h, z_i), Q2=(x_i, y_i - h, z_i)
    P3=(x_i, y_i, z_i + h), Q3=(x_i, y_i, z_i - h)

    (by mapping Pi and Qi to the grid using img_world2grid: the inverse of the
    grid-to-space transform of img). The displacement parameter h is of
    magnitude 0.5 (in physical space units), therefore the approximated partial
    derivatives are given by the difference between the image interpolated at
    Pi and Qi.

    Parameters
    ----------
    img : array, shape (S, R, C)
        the input volume whose gradient will be computed
    img_world2grid : array, shape (4, 4)
        the space-to-grid transform matrix associated to img
    img_spacing : array, shape (3,)
        the spacing between voxels (voxel size along each axis) of the input
    sample_points: array, shape (n, 3)
        list of points where the derivative will be evaluated
        (one point per row)
    out : array, shape (n, 3)
        the buffer in which to store the image gradient
    """
    cdef:
        int n = sample_points.shape[0]
        int i, in_flag
        double tmp
        double[:] dx = np.empty(shape=(3,), dtype=np.float64)
        double[:] h = np.empty(shape=(3,), dtype=np.float64)
        double[:] q = np.empty(shape=(3,), dtype=np.float64)
    with nogil:
        h[0] = 0.5 * img_spacing[0]
        h[1] = 0.5 * img_spacing[1]
        h[2] = 0.5 * img_spacing[2]
        for i in range(n):
            inside[i] = 1
            dx[0] = sample_points[i, 0]
            dx[1] = sample_points[i, 1]
            dx[2] = sample_points[i, 2]
            for p in range(3):
                # Compute coordinates of point dx on img's grid
                dx[p] = sample_points[i, p] - h[p]
                q[0] = _apply_affine_3d_x0(dx[0], dx[1], dx[2], 1,
                                           img_world2grid)
                q[1] = _apply_affine_3d_x1(dx[0], dx[1], dx[2], 1,
                                           img_world2grid)
                q[2] = _apply_affine_3d_x2(dx[0], dx[1], dx[2], 1,
                                           img_world2grid)
                # Interpolate img at q
                in_flag = _interpolate_scalar_3d[floating](img, q[0], q[1],
                    q[2], &out[i, p])
                if in_flag == 0:
                    out[i, p] = 0
                    inside[i] = 0
                    continue
                tmp = out[i, p]
                # Compute coordinates of point dx on img's grid
                dx[p] = sample_points[i, p] + h[p]
                q[0] = _apply_affine_3d_x0(dx[0], dx[1], dx[2], 1,
                                           img_world2grid)
                q[1] = _apply_affine_3d_x1(dx[0], dx[1], dx[2], 1,
                                           img_world2grid)
                q[2] = _apply_affine_3d_x2(dx[0], dx[1], dx[2], 1,
                                           img_world2grid)
                # Interpolate img at q
                in_flag = _interpolate_scalar_3d[floating](img, q[0], q[1], q[2],
                                                 &out[i, p])
                if in_flag == 0:
                    out[i, p] = 0
                    inside[i] = 0
                    continue
                out[i, p] = (out[i, p] - tmp) / img_spacing[p]
                dx[p] = sample_points[i, p]


def _gradient_2d(floating[:, :] img, double[:, :] img_world2grid,
                 double[:] img_spacing, double[:, :] out_grid2world,
                 floating[:, :, :] out, int[:, :] inside):
    r""" Gradient of a 2D image in physical space coordinates

    Each grid cell (i, j) in the sampling grid (determined by
    out.shape) is mapped to its corresponding physical point (x, y) by
    multiplying out_grid2world (its grid-to-space transform) by (i, j), then
    the image is interpolated, at

    P1=(x + h, y), Q1=(x - h, y)
    P2=(x, y + h), Q2=(x, y - h)

    (by mapping Pi and Qi to the grid using img_world2grid: the inverse of the
    grid-to-space transform of img). The displacement parameter h is of
    magnitude 0.5 (in physical space units), therefore the approximated partial
    derivatives are given by the difference between the image interpolated at
    Pi and Qi.

    Parameters
    ----------
    img : array, shape (R, C)
        the input image whose gradient will be computed
    img_world2grid : array, shape (3, 3)
        the space-to-grid transform matrix associated to img
    img_spacing : array, shape (2,)
        the spacing between pixels (pixel size along each axis) of the input
        image
    out_grid2world : array, shape (3, 3)
        the grid-to-space transform associated to the sampling grid
    out : array, shape (S', R', 2)
        the buffer in which to store the image gradient
    inside : array, shape (S', R')
        the buffer in which to store the flags indicating whether the sample
        point lies inside (=1) or outside (=0) the image grid
    """
    cdef:
        int nrows = out.shape[0]
        int ncols = out.shape[1]
        int i, j, k, in_flag
        double tmp
        double[:] x = np.empty(shape=(2,), dtype=np.float64)
        double[:] dx = np.empty(shape=(2,), dtype=np.float64)
        double[:] h = np.empty(shape=(2,), dtype=np.float64)
        double[:] q = np.empty(shape=(2,), dtype=np.float64)
    with nogil:
        h[0] = 0.5 * img_spacing[0]
        h[1] = 0.5 * img_spacing[1]
        for i in range(nrows):
            for j in range(ncols):
                inside[i, j] = 1
                # Compute coordinates of index (i, j) in physical space
                x[0] = _apply_affine_2d_x0(i, j, 1, out_grid2world)
                x[1] = _apply_affine_2d_x1(i, j, 1, out_grid2world)
                dx[:] = x[:]
                for p in range(2):
                    # Compute coordinates of point dx on img's grid
                    dx[p] = x[p] - h[p]
                    q[0] = _apply_affine_2d_x0(dx[0], dx[1], 1, img_world2grid)
                    q[1] = _apply_affine_2d_x1(dx[0], dx[1], 1, img_world2grid)
                    # Interpolate img at q
                    in_flag = _interpolate_scalar_2d[floating](img, q[0],
                        q[1], &out[i, j, p])
                    if in_flag == 0:
                        out[i, j, p] = 0
                        inside[i, j] = 0
                        continue
                    tmp = out[i, j, p]
                    # Compute coordinates of point dx on img's grid
                    dx[p] = x[p] + h[p]
                    q[0] = _apply_affine_2d_x0(dx[0], dx[1], 1, img_world2grid)
                    q[1] = _apply_affine_2d_x1(dx[0], dx[1], 1, img_world2grid)
                    # Interpolate img at q
                    in_flag = _interpolate_scalar_2d[floating](img, q[0],
                        q[1], &out[i, j, p])
                    if in_flag == 0:
                        out[i, j, p] = 0
                        inside[i, j] = 0
                        continue
                    out[i, j, p] = (out[i, j, p] - tmp) / img_spacing[p]
                    dx[p] = x[p]


def _sparse_gradient_2d(floating[:, :] img, double[:, :] img_world2grid,
                        double[:] img_spacing,
                        double[:, :] sample_points,
                        floating[:, :] out, int[:] inside):
    r""" Gradient of a 2D image evaluated at a set of points in physical space

    For each row (x_i, y_i) in sample_points, the image is interpolated at

    P1=(x_i + h, y_i), Q1=(x_i - h, y_i)
    P2=(x_i, y_i + h), Q2=(x_i, y_i - h)

    (by mapping Pi and Qi to the grid using img_world2grid: the inverse of the
    grid-to-space transform of img). The displacement parameter h is of
    magnitude 0.5 (in physical space units), therefore the approximated partial
    derivatives are given by the difference between the image interpolated at
    Pi and Qi.

    Parameters
    ----------
    img : array, shape (R, C)
        the input volume whose gradient will be computed
    img_world2grid : array, shape (3, 3)
        the space-to-grid transform matrix associated to img
    img_spacing : array, shape (2,)
        the spacing between pixels (pixel size along each axis) of the input
    sample_points: array, shape (n, 2)
        list of points where the derivative will be evaluated
        (one point per row)
    out : array, shape (n, 2)
        the buffer in which to store the image gradient
    inside : array, shape (n,)
        the buffer in which to store the flags indicating whether the sample
        point lies inside (=1) or outside (=0) the image grid
    """
    cdef:
        int n = sample_points.shape[0]
        int i, in_flag
        double tmp
        double[:] dx = np.empty(shape=(2,), dtype=np.float64)
        double[:] h = np.empty(shape=(2,), dtype=np.float64)
        double[:] q = np.empty(shape=(2,), dtype=np.float64)
    with nogil:
        h[0] = 0.5 * img_spacing[0]
        h[1] = 0.5 * img_spacing[1]
        for i in range(n):
            inside[i] = 1
            dx[0] = sample_points[i, 0]
            dx[1] = sample_points[i, 1]
            for p in range(2):
                # Compute coordinates of point dx on img's grid
                dx[p] = sample_points[i, p] - h[p]
                q[0] = _apply_affine_2d_x0(dx[0], dx[1], 1, img_world2grid)
                q[1] = _apply_affine_2d_x1(dx[0], dx[1], 1, img_world2grid)
                # Interpolate img at q
                in_flag = _interpolate_scalar_2d[floating](img, q[0], q[1],
                                                           &out[i, p])
                if in_flag == 0:
                    out[i, p] = 0
                    inside[i] = 0
                    continue
                tmp = out[i, p]
                # Compute coordinates of point dx on img's grid
                dx[p] = sample_points[i, p] + h[p]
                q[0] = _apply_affine_2d_x0(dx[0], dx[1], 1, img_world2grid)
                q[1] = _apply_affine_2d_x1(dx[0], dx[1], 1, img_world2grid)
                # Interpolate img at q
                in_flag = _interpolate_scalar_2d[floating](img, q[0], q[1],
                                                           &out[i, p])
                if in_flag == 0:
                    out[i, p] = 0
                    inside[i] = 0
                    continue
                out[i, p] = (out[i, p] - tmp) / img_spacing[p]
                dx[p] = sample_points[i, p]


def gradient(img, img_world2grid, img_spacing, out_shape,
             out_grid2world):
    r""" Gradient of an image in physical space

    Parameters
    ----------
    img : 2D or 3D array, shape (R, C) or (S, R, C)
        the input image whose gradient will be computed
    img_world2grid : array, shape (dim+1, dim+1)
        the space-to-grid transform matrix associated to img
    img_spacing : array, shape (dim,)
        the spacing between voxels (voxel size along each axis) of the input
        image
    out_shape : array, shape (dim,)
        the number of (slices), rows and columns of the sampling grid
    out_grid2world : array, shape (dim+1, dim+1)
        the grid-to-space transform associated to the sampling grid

    Returns
    -------
    out : array, shape (R', C', 2) or (S', R', C', 3)
        the buffer in which to store the image gradient, where
        (S'), R', C' are given by out_shape
    """
    dim = len(img.shape)
    if not is_valid_affine(img_world2grid, dim):
        raise ValueError("Invalid image affine transform")
    if not is_valid_affine(out_grid2world, dim):
        raise ValueError("Invalid sampling grid affine transform")
    if len(img_spacing) < dim:
        raise ValueError("Invalid spacings")
    ftype = img.dtype.type
    out = np.empty(tuple(out_shape)+(dim,), dtype=ftype)
    inside = np.empty(tuple(out_shape), dtype=np.int32)
    # Select joint density gradient 2D or 3D
    if dim == 2:
        jd_grad = _gradient_2d
    elif dim == 3:
        jd_grad = _gradient_3d
    else:
        raise ValueError('Undefined gradient for image dimension %d' % (dim,))
    if img_world2grid.dtype != np.float64:
        img_world2grid = img_world2grid.astype(np.float64)
    if img_spacing.dtype != np.float64:
        img_spacing = img_spacing.astype(np.float64)
    if out_grid2world.dtype != np.float64:
        out_grid2world = out_grid2world.astype(np.float64)
    jd_grad(img, img_world2grid, img_spacing, out_grid2world, out, inside)
    return np.asarray(out), np.asarray(inside)


def sparse_gradient(img, img_world2grid, img_spacing, sample_points):
    r""" Gradient of an image in physical space

    Parameters
    ----------
    img : 2D or 3D array, shape (R, C) or (S, R, C)
        the input image whose gradient will be computed
    img_world2grid : array, shape (dim+1, dim+1)
        the space-to-grid transform matrix associated to img
    img_spacing : array, shape (dim,)
        the spacing between voxels (voxel size along each axis) of the input
        image
    sample_points: array, shape (n, dim)
        list of points where the derivative will be evaluated
        (one point per row)

    Returns
    -------
    out : array, shape (n, dim)
        the gradient at each point stored at its corresponding row
    """
    dim = len(img.shape)
    if not is_valid_affine(img_world2grid, dim):
        raise ValueError("Invalid affine transform matrix")
    if len(img_spacing) < dim:
        raise ValueError("Invalid spacings")

    ftype = img.dtype.type
    n = sample_points.shape[0]
    out = np.empty(shape=(n, dim), dtype=ftype)
    inside = np.empty(shape=(n,), dtype=np.int32)
    # Select joint density gradient 2D or 3D
    if dim == 2:
        jd_grad = _sparse_gradient_2d
    else:
        jd_grad = _sparse_gradient_3d
    if img_world2grid.dtype != np.float64:
        img_world2grid = img_world2grid.astype(np.float64)
    if img_spacing.dtype != np.float64:
        img_spacing = img_spacing.astype(np.float64)
    jd_grad(img, img_world2grid, img_spacing, sample_points, out, inside)
    return np.asarray(out), np.asarray(inside)



import itertools
import numpy as np
from numpy.testing import (assert_array_equal, assert_equal, assert_raises,
                           run_module_suite)

from dipy.segment.clustering import QuickBundlesX, QuickBundles, qbx_and_merge
from dipy.segment.featurespeed import ResampleFeature
from dipy.segment.metric import AveragePointwiseEuclideanMetric
from dipy.segment.metric import MinimumAverageDirectFlipMetric
from dipy.tracking.streamline import set_number_of_points
from dipy.tracking.streamline import Streamlines


def straight_bundle(nb_streamlines=1, nb_pts=30, step_size=1,
                    radius=1, rng=np.random.RandomState(42)):
    bundle = []

    bundle_length = step_size * nb_pts

    Z = -np.linspace(0, bundle_length, nb_pts)
    for k in range(nb_streamlines):
        theta = rng.rand() * (2*np.pi)
        r = radius * rng.rand()

        Xk = np.ones(nb_pts) * (r * np.cos(theta))
        Yk = np.ones(nb_pts) * (r * np.sin(theta))
        Zk = Z.copy()

        bundle.append(np.c_[Xk, Yk, Zk])

    return bundle


def bearing_bundles(nb_balls=6, bearing_radius=2):
    bundles = []

    for theta in np.linspace(0, 2*np.pi, nb_balls, endpoint=False):
        x = bearing_radius * np.cos(theta)
        y = bearing_radius * np.sin(theta)

        bundle = np.array(straight_bundle(nb_streamlines=100))
        bundle += (x, y, 0)
        bundles.append(bundle)

    return bundles


def streamlines_in_circle(nb_streamlines=1, nb_pts=30, step_size=1,
                          radius=1):
    bundle = []

    bundle_length = step_size * nb_pts

    Z = np.linspace(0, bundle_length, nb_pts)
    for theta in np.linspace(0, 2*np.pi, nb_streamlines, endpoint=False):
        Xk = np.ones(nb_pts) * (radius * np.cos(theta))
        Yk = np.ones(nb_pts) * (radius * np.sin(theta))
        Zk = Z.copy()

        bundle.append(np.c_[Xk, Yk, Zk])

    return bundle


def streamlines_parallel(nb_streamlines=1, nb_pts=30, step_size=1,
                         delta=1):
    bundle = []

    bundle_length = step_size * nb_pts

    Z = np.linspace(0, bundle_length, nb_pts)
    for x in delta*np.arange(0, nb_streamlines):
        Xk = np.ones(nb_pts) * x
        Yk = np.zeros(nb_pts)
        Zk = Z.copy()

        bundle.append(np.c_[Xk, Yk, Zk])

    return bundle


def simulated_bundle(no_streamlines=10, waves=False, no_pts=12):
    t = np.linspace(-10, 10, 200)
    # parallel waves or parallel lines
    bundle = []
    for i in np.linspace(-5, 5, no_streamlines):
        if waves:
            pts = np.vstack((np.cos(t), t, i * np.ones(t.shape))).T
        else:
            pts = np.vstack((np.zeros(t.shape), t, i * np.ones(t.shape))).T
        pts = set_number_of_points(pts, no_pts)
        bundle.append(pts)

    return bundle


def test_3D_points():

    points = np.array([[[1, 0, 0]],
                       [[3, 0, 0]],
                       [[2, 0, 0]],
                       [[5, 0, 0]],
                       [[5.5, 0, 0]]], dtype="f4")

    thresholds = [4, 2, 1]
    metric = AveragePointwiseEuclideanMetric()
    qbx = QuickBundlesX(thresholds, metric=metric)
    tree = qbx.cluster(points)
    clusters_2 = tree.get_clusters(2)
    assert_array_equal(clusters_2.clusters_sizes(), [3, 2])
    clusters_0 = tree.get_clusters(0)
    assert_array_equal(clusters_0.clusters_sizes(), [5])


def test_3D_segments():
    points = np.array([[[1, 0, 0],
                        [1, 1, 0]],
                       [[3, 1, 0],
                        [3, 0, 0]],
                       [[2, 0, 0],
                        [2, 1, 0]],
                       [[5, 1, 0],
                        [5, 0, 0]],
                       [[5.5, 0, 0],
                        [5.5, 1, 0]]], dtype="f4")

    thresholds = [4, 2, 1]

    feature = ResampleFeature(nb_points=20)
    metric = AveragePointwiseEuclideanMetric(feature)
    qbx = QuickBundlesX(thresholds, metric=metric)
    tree = qbx.cluster(points)
    clusters_0 = tree.get_clusters(0)
    clusters_1 = tree.get_clusters(1)
    clusters_2 = tree.get_clusters(2)

    assert_equal(len(clusters_0.centroids), len(clusters_1.centroids))
    assert_equal(len(clusters_2.centroids) > len(clusters_1.centroids), True)

    assert_array_equal(clusters_2[1].indices, np.array([3, 4], dtype=np.int32))


def test_with_simulated_bundles():

    streamlines = simulated_bundle(3, False, 2)
    thresholds = [10, 3, 1]
    qbx_class = QuickBundlesX(thresholds)
    tree = qbx_class.cluster(streamlines)
    for level in range(len(thresholds) + 1):
        clusters = tree.get_clusters(level)

    assert_equal(tree.leaves[0].indices[0], 0)
    assert_equal(tree.leaves[2][0], 2)
    clusters.refdata = streamlines

    assert_array_equal(clusters[0][0],
                       np.array([[0., -10., -5.],
                                 [0., 10., -5.]]))


def test_with_simulated_bundles2():

    # Generate synthetic streamlines
    bundles = bearing_bundles(4, 2)
    bundles.append(straight_bundle(1))
    streamlines = list(itertools.chain(*bundles))

    thresholds = [10, 2, 1]
    qbx_class = QuickBundlesX(thresholds)
    tree = qbx_class.cluster(streamlines)
    # By default `refdata` refers to data being clustered.
    assert_equal(tree.refdata, streamlines)


def test_circle_parallel_fornix():

    circle = streamlines_in_circle(100, step_size=2)

    parallel = streamlines_parallel(100)

    thresholds = [1, 0.1]

    qbx_class = QuickBundlesX(thresholds)
    tree = qbx_class.cluster(circle)

    clusters = tree.get_clusters(0)
    assert_equal(len(clusters), 1)

    clusters = tree.get_clusters(1)
    assert_equal(len(clusters), 3)

    clusters = tree.get_clusters(2)
    assert_equal(len(clusters), 34)

    thresholds = [.5]

    qbx_class = QuickBundlesX(thresholds)
    tree = qbx_class.cluster(parallel)

    clusters = tree.get_clusters(0)
    assert_equal(len(clusters), 1)

    clusters = tree.get_clusters(1)
    assert_equal(len(clusters), 100)


def test_raise_mdf():

    thresholds = [1, 0.1]

    metric = MinimumAverageDirectFlipMetric()

    assert_raises(ValueError, QuickBundlesX, thresholds, metric=metric)
    assert_raises(ValueError, QuickBundles, thresholds[1], metric=metric)


def test_qbx_and_merge():

    # Generate synthetic streamlines
    bundles = bearing_bundles(4, 2)
    bundles.append(straight_bundle(1))

    streamlines = Streamlines(list(itertools.chain(*bundles)))

    thresholds = [10, 2, 1]

    rng = np.random.RandomState(seed=42)
    qbxm_centroids = qbx_and_merge(streamlines, thresholds, rng=rng).centroids

    qbx = QuickBundlesX(thresholds)
    tree = qbx.cluster(streamlines)
    qbx_centroids = tree.get_clusters(3).centroids

    assert_equal(len(qbx_centroids) > len(qbxm_centroids), True)


if __name__ == '__main__':
    run_module_suite()



# Init for tracking module
""" Tracking objects """

from nibabel.streamlines import ArraySequence as Streamlines

# Test callable
from numpy.testing import Tester
test = Tester().test
bench = Tester().bench
del Tester



from warnings import warn

import dipy
import dipy.reconst.eudx_direction_getter


def EuDXDirectionGetter(*args, **kwargs):
    warn(DeprecationWarning(
            "class 'dipy.reconst.peak_direction_getter.EuDXDirectionGetter'"
            " is deprecated since version 1.2.0, use class"
            " 'dipy.reconst.eudx_direction_getter.EuDXDirectionGetter'"
            " instead"))
    return dipy.reconst.eudx_direction_getter.EuDXDirectionGetter(*args,
                                                                  **kwargs)



"""
==============================================
Denoise images using Non-Local Means (NLMEANS)
==============================================

Using the non-local means filter [Coupe08]_ and [Coupe11]_ and  you can denoise
3D or 4D images and boost the SNR of your datasets. You can also decide between
modeling the noise as Gaussian or Rician (default).

"""

import numpy as np
import nibabel as nib
import matplotlib.pyplot as plt
from time import time
from dipy.denoise.nlmeans import nlmeans
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.data import get_fnames
from dipy.io.image import load_nifti


dwi_fname, dwi_bval_fname, dwi_bvec_fname = get_fnames('sherbrooke_3shell')
data, affine = load_nifti(dwi_fname)

mask = data[..., 0] > 80

# We select only one volume for the example to run quickly.
data = data[..., 1]

print("vol size", data.shape)

# lets create a noisy data with Gaussian data

"""
In order to call ``non_local_means`` first you need to estimate the standard
deviation of the noise. We use N=4 since the Sherbrooke dataset was acquired
on a 1.5T Siemens scanner with a 4 array head coil.
"""

sigma = estimate_sigma(data, N=4)

t = time()

"""
Calling the main function ``non_local_means``
"""

t = time()

den = nlmeans(data, sigma=sigma, mask=mask, patch_radius=1,
              block_radius=1, rician=True)

print("total time", time() - t)
"""
Let us plot the axial slice of the denoised output
"""

axial_middle = data.shape[2] // 2

before = data[:, :, axial_middle].T
after = den[:, :, axial_middle].T

difference = np.abs(after.astype(np.float64) - before.astype(np.float64))

difference[~mask[:, :, axial_middle].T] = 0


fig, ax = plt.subplots(1, 3)
ax[0].imshow(before, cmap='gray', origin='lower')
ax[0].set_title('before')
ax[1].imshow(after, cmap='gray', origin='lower')
ax[1].set_title('after')
ax[2].imshow(difference, cmap='gray', origin='lower')
ax[2].set_title('difference')

plt.savefig('denoised.png', bbox_inches='tight')


"""
.. figure:: denoised.png
   :align: center

   **Showing axial slice before (left) and after (right) NLMEANS denoising**
"""

nib.save(nib.Nifti1Image(den, affine), 'denoised.nii.gz')

"""
An improved version of non-local means denoising is adaptive soft coefficient
matching, please refer to :ref:`example_denoise_ascm` for more details.

References
----------

.. [Coupe08] P. Coupe, P. Yger, S. Prima, P. Hellier, C. Kervrann, C. Barillot,
   "An Optimized Blockwise Non Local Means Denoising Filter for 3D Magnetic
   Resonance Images", IEEE Transactions on Medical Imaging, 27(4):425-441, 2008

.. [Coupe11] Pierrick Coupe, Jose Manjon, Montserrat Robles, Louis Collins.
    "Adaptive Multiresolution Non-Local Means Filter for 3D MR Image Denoising"
    IET Image Processing, Institution of Engineering and Technology, 2011

.. include:: ../links_names.inc
"""



cimport numpy as np
import numpy as np

from dipy.data import default_sphere
from dipy.direction.closest_peak_direction_getter cimport (closest_peak,
                                                           BasePmfDirectionGetter)
from dipy.direction.pmf import BootPmfGen


cdef class BootDirectionGetter(BasePmfDirectionGetter):

    cdef:
        int max_attempts

    def __init__(self, pmfgen, maxangle, sphere=default_sphere,
                 max_attempts=5, **kwargs):
        if max_attempts < 1:
             raise ValueError("max_attempts must be greater than 0.")
        self.max_attempts = max_attempts
        BasePmfDirectionGetter.__init__(self, pmfgen, maxangle, sphere, **kwargs)

    @classmethod
    def from_data(klass, data, model, max_angle, sphere=default_sphere,
                  sh_order=0, max_attempts=5, **kwargs):
        """Create a BootDirectionGetter using HARDI data and an ODF type model

        Parameters
        ----------
        data : ndarray, float, (..., N)
            Diffusion MRI data with N volumes.
        model : dipy diffusion model
            Must provide fit with odf method.
        max_angle : float (0, 90)
            Maximum angle between tract segments. This angle can be more
            generous (larger) than values typically used with probabilistic
            direction getters.
        sphere : Sphere
            The sphere used to sample the diffusion ODF.
        sh_order : even int
            The order of the SH "model" used to estimate bootstrap residuals.
        max_attempts : int
            Max number of bootstrap samples used to find tracking direction
            before giving up.
        pmf_threshold : float
            Threshold for ODF functions.
        relative_peak_threshold : float in [0., 1.]
            Relative threshold for excluding ODF peaks.
        min_separation_angle : float in [0, 90]
            Angular threshold for excluding ODF peaks.

        """
        boot_gen = BootPmfGen(np.asarray(data, dtype=float), model, sphere,
                              sh_order=sh_order)
        return klass(boot_gen, max_angle, sphere, max_attempts, **kwargs)

    cdef int get_direction_c(self, double* point, double* direction):
        """Attempt direction getting on a few bootstrap samples.

        Returns
        -------
        status : int
            Returns 0 `direction` was updated with a new tracking direction, or
            1 otherwise.
        """
        cdef:
            double[:] pmf,
            np.ndarray[np.float_t, ndim=2] peaks

        for _ in range(self.max_attempts):
            pmf = self._get_pmf(point)
            peaks = self._get_peak_directions(pmf)
            if len(peaks) > 0:
                return closest_peak(peaks, direction, self.cos_similarity)
        return 1



import random
from collections.abc import Iterable

import numpy as np

from dipy.tracking.localtrack import local_tracker, pft_tracker
from dipy.tracking.stopping_criterion import (AnatomicalStoppingCriterion,
                                              StreamlineStatus)
from dipy.tracking import utils


class LocalTracking(object):

    @staticmethod
    def _get_voxel_size(affine):
        """Computes the voxel sizes of an image from the affine.

        Checks that the affine does not have any shear because local_tracker
        assumes that the data is sampled on a regular grid.

        """
        lin = affine[:3, :3]
        dotlin = np.dot(lin.T, lin)
        # Check that the affine is well behaved
        if not np.allclose(np.triu(dotlin, 1), 0., atol=1e-5):
            msg = ("The affine provided seems to contain shearing, data must "
                   "be acquired or interpolated on a regular grid to be used "
                   "with `LocalTracking`.")
            raise ValueError(msg)
        return np.sqrt(dotlin.diagonal())

    def __init__(self, direction_getter, stopping_criterion, seeds, affine,
                 step_size, max_cross=None, maxlen=500, fixedstep=True,
                 return_all=True, random_seed=None, save_seeds=False):
        """Creates streamlines by using local fiber-tracking.

        Parameters
        ----------
        direction_getter : instance of DirectionGetter
            Used to get directions for fiber tracking.
        stopping_criterion : instance of StoppingCriterion
            Identifies endpoints and invalid points to inform tracking.
        seeds : array (N, 3)
            Points to seed the tracking. Seed points should be given in point
            space of the track (see ``affine``).
        affine : array (4, 4)
            Coordinate space for the streamline point with respect to voxel
            indices of input data. This affine can contain scaling, rotational,
            and translational components but should not contain any shearing.
            An identity matrix can be used to generate streamlines in "voxel
            coordinates" as long as isotropic voxels were used to acquire the
            data.
        step_size : float
            Step size used for tracking.
        max_cross : int or None
            The maximum number of direction to track from each seed in crossing
            voxels. By default all initial directions are tracked.
        maxlen : int
            Maximum number of steps to track from seed. Used to prevent
            infinite loops.
        fixedstep : bool
            If true, a fixed stepsize is used, otherwise a variable step size
            is used.
        return_all : bool
            If true, return all generated streamlines, otherwise only
            streamlines reaching end points or exiting the image.
        random_seed : int
            The seed for the random seed generator (numpy.random.seed and
            random.seed).
        save_seeds : bool
            If True, return seeds alongside streamlines
        """

        self.direction_getter = direction_getter
        self.stopping_criterion = stopping_criterion
        self.seeds = seeds
        if affine.shape != (4, 4):
            raise ValueError("affine should be a (4, 4) array.")
        if step_size <= 0:
            raise ValueError("step_size must be greater than 0.")
        if maxlen < 1:
            raise ValueError("maxlen must be greater than 0.")
        if not isinstance(seeds, Iterable):
            raise ValueError("seeds should be (N,3) array.")

        self.affine = affine
        self._voxel_size = np.ascontiguousarray(self._get_voxel_size(affine),
                                                dtype=float)
        self.step_size = step_size
        self.fixed_stepsize = fixedstep
        self.max_cross = max_cross
        self.max_length = maxlen
        self.return_all = return_all
        self.random_seed = random_seed
        self.save_seeds = save_seeds

    def _tracker(self, seed, first_step, streamline):
        return local_tracker(self.direction_getter,
                             self.stopping_criterion,
                             seed,
                             first_step,
                             self._voxel_size,
                             streamline,
                             self.step_size,
                             self.fixed_stepsize)

    def __iter__(self):
        # Make tracks, move them to point space and return
        track = self._generate_streamlines()

        return utils.transform_tracking_output(track, self.affine,
                                               save_seeds=self.save_seeds)

    def _generate_streamlines(self):
        """A streamline generator"""

        # Get inverse transform (lin/offset) for seeds
        inv_A = np.linalg.inv(self.affine)
        lin = inv_A[:3, :3]
        offset = inv_A[:3, 3]

        F = np.empty((self.max_length + 1, 3), dtype=float)
        B = F.copy()
        for s in self.seeds:
            s = np.dot(lin, s) + offset
            # Set the random seed in numpy and random
            if self.random_seed is not None:
                s_random_seed = hash(np.abs((np.sum(s)) + self.random_seed)) \
                    % (2**32 - 1)
                random.seed(s_random_seed)
                np.random.seed(s_random_seed)
            directions = self.direction_getter.initial_direction(s)
            if directions.size == 0 and self.return_all:
                # only the seed position
                if self.save_seeds:
                    yield [s], s
                else:
                    yield [s]
            directions = directions[:self.max_cross]
            for first_step in directions:
                stepsF, stream_status = self._tracker(s, first_step, F)
                if not (self.return_all or
                        stream_status == StreamlineStatus.ENDPOINT or
                        stream_status == StreamlineStatus.OUTSIDEIMAGE):
                    continue
                first_step = -first_step
                stepsB, stream_status = self._tracker(s, first_step, B)
                if not (self.return_all or
                        stream_status == StreamlineStatus.ENDPOINT or
                        stream_status == StreamlineStatus.OUTSIDEIMAGE):
                    continue
                if stepsB == 1:
                    streamline = F[:stepsF].copy()
                else:
                    parts = (B[stepsB - 1:0:-1], F[:stepsF])
                    streamline = np.concatenate(parts, axis=0)

                # move to the next streamline if only the seed position
                # and not return all
                if len(streamline) > 1 or self.return_all:
                    if self.save_seeds:
                        yield streamline, s
                    else:
                        yield streamline


class ParticleFilteringTracking(LocalTracking):

    def __init__(self, direction_getter, stopping_criterion, seeds, affine,
                 step_size, max_cross=None, maxlen=500,
                 pft_back_tracking_dist=2, pft_front_tracking_dist=1,
                 pft_max_trial=20, particle_count=15, return_all=True,
                 random_seed=None, save_seeds=False):
        r"""A streamline generator using the particle filtering tractography
        method [1]_.

        Parameters
        ----------
        direction_getter : instance of ProbabilisticDirectionGetter
            Used to get directions for fiber tracking.
        stopping_criterion : instance of AnatomicalStoppingCriterion
            Identifies endpoints and invalid points to inform tracking.
        seeds : array (N, 3)
            Points to seed the tracking. Seed points should be given in point
            space of the track (see ``affine``).
        affine : array (4, 4)
            Coordinate space for the streamline point with respect to voxel
            indices of input data. This affine can contain scaling, rotational,
            and translational components but should not contain any shearing.
            An identity matrix can be used to generate streamlines in "voxel
            coordinates" as long as isotropic voxels were used to acquire the
            data.
        step_size : float
            Step size used for tracking.
        max_cross : int or None
            The maximum number of direction to track from each seed in crossing
            voxels. By default all initial directions are tracked.
        maxlen : int
            Maximum number of steps to track from seed. Used to prevent
            infinite loops.
        pft_back_tracking_dist : float
            Distance in mm to back track before starting the particle filtering
            tractography. The total particle filtering tractography distance is
            equal to back_tracking_dist + front_tracking_dist.
            By default this is set to 2 mm.
        pft_front_tracking_dist : float
            Distance in mm to run the particle filtering tractography after the
            the back track distance. The total particle filtering tractography
            distance is equal to back_tracking_dist + front_tracking_dist. By
            default this is set to 1 mm.
        pft_max_trial : int
            Maximum number of trial for the particle filtering tractography
            (Prevents infinite loops).
        particle_count : int
            Number of particles to use in the particle filter.
        return_all : bool
            If true, return all generated streamlines, otherwise only
            streamlines reaching end points or exiting the image.
        random_seed : int
            The seed for the random seed generator (numpy.random.seed and
            random.seed).
        save_seeds : bool
            If True, return seeds alongside streamlines


        References
        ----------
        .. [1] Girard, G., Whittingstall, K., Deriche, R., & Descoteaux, M.
               Towards quantitative connectivity analysis: reducing
               tractography biases. NeuroImage, 98, 266-278, 2014.
        """

        if not isinstance(stopping_criterion, AnatomicalStoppingCriterion):
            raise ValueError("expecting AnatomicalStoppingCriterion")

        self.pft_max_nbr_back_steps = int(np.ceil(pft_back_tracking_dist
                                                  / step_size))
        self.pft_max_nbr_front_steps = int(np.ceil(pft_front_tracking_dist
                                                   / step_size))
        pft_max_steps = (self.pft_max_nbr_back_steps +
                         self.pft_max_nbr_front_steps)

        if (self.pft_max_nbr_front_steps < 0
                or self.pft_max_nbr_back_steps < 0
                or pft_max_steps < 1):
            raise ValueError("The number of PFT steps must be greater than 0.")

        if particle_count <= 0:
            raise ValueError("The particle count must be greater than 0.")

        self.directions = np.empty((maxlen + 1, 3), dtype=float)

        self.pft_max_trial = pft_max_trial
        self.particle_count = particle_count
        self.particle_paths = np.empty((2, self.particle_count,
                                        pft_max_steps + 1, 3),
                                       dtype=float)
        self.particle_weights = np.empty(self.particle_count, dtype=float)
        self.particle_dirs = np.empty((2, self.particle_count,
                                       pft_max_steps + 1, 3), dtype=float)
        self.particle_steps = np.empty((2, self.particle_count), dtype=int)
        self.particle_stream_statuses = np.empty((2, self.particle_count),
                                                 dtype=int)
        super(ParticleFilteringTracking, self).__init__(direction_getter,
                                                        stopping_criterion,
                                                        seeds,
                                                        affine,
                                                        step_size,
                                                        max_cross,
                                                        maxlen,
                                                        True,
                                                        return_all,
                                                        random_seed,
                                                        save_seeds)

    def _tracker(self, seed, first_step, streamline):
        return pft_tracker(self.direction_getter,
                           self.stopping_criterion,
                           seed,
                           first_step,
                           self._voxel_size,
                           streamline,
                           self.directions,
                           self.step_size,
                           self.pft_max_nbr_back_steps,
                           self.pft_max_nbr_front_steps,
                           self.pft_max_trial,
                           self.particle_count,
                           self.particle_paths,
                           self.particle_dirs,
                           self.particle_weights,
                           self.particle_steps,
                           self.particle_stream_statuses)



"""
=========================================
Tractography Clustering with QuickBundles
=========================================

This example explains how we can use QuickBundles [Garyfallidis12]_ to
simplify/cluster streamlines.

First import the necessary modules.
"""

import numpy as np
from dipy.io.streamline import load_tractogram
from dipy.tracking.streamline import Streamlines
from dipy.segment.clustering import QuickBundles
from dipy.io.pickles import save_pickle
from dipy.data import get_fnames
from dipy.viz import window, actor

"""
For educational purposes we will try to cluster a small streamline bundle known
from neuroanatomy as the fornix.
"""

fname = get_fnames('fornix')

"""
Load fornix streamlines.
"""

fornix = load_tractogram(fname, 'same', bbox_valid_check=False)
streamlines = fornix.streamlines

"""
Perform QuickBundles clustering using the MDF metric and a 10mm distance
threshold. Keep in mind that since the MDF metric requires streamlines to have
the same number of points, the clustering algorithm will internally use a
representation of streamlines that have been automatically downsampled/upsampled
so they have only 12 points (To set manually the number of points,
see :ref:`clustering-examples-ResampleFeature`).
"""

qb = QuickBundles(threshold=10.)
clusters = qb.cluster(streamlines)

"""
`clusters` is a `ClusterMap` object which contains attributes that
provide information about the clustering result.
"""

print("Nb. clusters:", len(clusters))
print("Cluster sizes:", map(len, clusters))
print("Small clusters:", clusters < 10)
print("Streamlines indices of the first cluster:\n", clusters[0].indices)
print("Centroid of the last cluster:\n", clusters[-1].centroid)

"""

::

    Nb. clusters: 4

    Cluster sizes: [64, 191, 47, 1]

    Small clusters: array([False, False, False, True], dtype=bool)

    Streamlines indices of the first cluster:
    [0, 7, 8, 10, 11, 12, 13, 14, 15, 18, 26, 30, 33, 35, 41, 65, 66, 85, 100,
     101, 105, 115, 116, 119, 122, 123, 124, 125, 126, 128, 129, 135, 139, 142,
     143, 144, 148, 151, 159, 167, 175, 180, 181, 185, 200, 208, 210, 224, 237,
     246, 249, 251, 256, 267, 270, 280, 284, 293, 296, 297, 299]

    Centroid of the last cluster:
    array([[  84.83773804,  117.92590332,   77.32278442],
           [  86.10850525,  115.84362793,   81.91885376],
           [  86.40357208,  112.25676727,   85.72930145],
           [  86.48336792,  107.60327911,   88.13782501],
           [  86.23897552,  102.5100708 ,   89.29447174],
           [  85.04563904,   97.46020508,   88.54240417],
           [  82.60240173,   93.14851379,   86.84208679],
           [  78.98937225,   89.57682037,   85.63652039],
           [  74.72344208,   86.60827637,   84.9391861 ],
           [  70.40846252,   85.15874481,   82.4484024 ],
           [  66.74534607,   86.00262451,   78.82582092],
           [  64.02451324,   88.43942261,   75.0697403 ]], dtype=float32)


`clusters` has also attributes like `centroids` (cluster representatives), and
methods like `add`, `remove`, and `clear` to modify the clustering result.

Lets first show the initial dataset.
"""

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, window.colors.white))
window.record(scene, out_path='fornix_initial.png', size=(600, 600))
if interactive:
    window.show(scene)

"""
.. figure:: fornix_initial.png
   :align: center

   Initial Fornix dataset.

Show the centroids of the fornix after clustering (with random colors):
"""

colormap = actor.create_colormap(np.arange(len(clusters)))

scene.clear()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, window.colors.white, opacity=0.05))
scene.add(actor.streamtube(clusters.centroids, colormap, linewidth=0.4))
window.record(scene, out_path='fornix_centroids.png', size=(600, 600))
if interactive:
    window.show(scene)

"""
.. figure:: fornix_centroids.png
   :align: center

   Showing the different QuickBundles centroids with random colors.

Show the labeled fornix (colors from centroids).
"""

colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

scene.clear()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, colormap_full))
window.record(scene, out_path='fornix_clusters.png', size=(600, 600))
if interactive:
    window.show(scene)

"""
.. figure:: fornix_clusters.png
   :align: center

   Showing the different clusters.

It is also possible to save the complete `ClusterMap` object with pickling.
"""

save_pickle('QB.pkl', clusters)

"""
Finally, here is a video of QuickBundles applied on a larger dataset.

.. raw:: html

    <iframe width="420" height="315" src="http://www.youtube.com/embed/kstL7KKqu94" frameborder="0" allowfullscreen></iframe>

.. include:: ../links_names.inc

References
----------

.. [Garyfallidis12] Garyfallidis E. et al., QuickBundles a method for
                    tractography simplification, Frontiers in Neuroscience, vol
                    6, no 175, 2012.

"""



"""

.. _reconst-csd:

=======================================================
Reconstruction with Constrained Spherical Deconvolution
=======================================================

This example shows how to use Constrained Spherical Deconvolution (CSD)
introduced by Tournier et al. [Tournier2007]_.

This method is mainly useful with datasets with gradient directions acquired on
a spherical grid.

The basic idea with this method is that if we could estimate the response
function of a single fiber then we could deconvolve the measured signal and
obtain the underlying fiber distribution.

In this way, the reconstruction of the fiber orientation distribution function
(fODF) in CSD involves two steps:
    1. Estimation of the fiber response function
    2. Use the response function to reconstruct the fODF

Let's first load the data. We will use a dataset with 10 b0s and 150 non-b0s
with b-value 2000.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, default_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames('stanford_hardi')

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs)

"""
You can verify the b-values of the dataset by looking at the attribute
``gtab.bvals``. Now that a datasets with multiple gradient directions is
loaded, we can proceed with the two steps of CSD.

Step 1. Estimation of the fiber response function
=================================================

There are many strategies to estimate the fiber response function. Here two
different strategies are presented.

**Strategy 1 - response function estimates from a local brain region**
One simple way to estimate the fiber response function is to look for regions
of the brain where it is known that there are single coherent fiber
populations. For example, if we use a ROI at the center of the brain, we will
find single fibers from the corpus callosum. The ``auto_response_ssst``
function will calculate FA for a cuboid ROI of radii equal to ``roi_radii`` in
the center of the volume and return the response function estimated in that
region for the voxels with FA higher than 0.7.
"""

from dipy.reconst.csdeconv import (auto_response_ssst,
                                   mask_for_response_ssst,
                                   response_from_mask_ssst)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

"""
Note that the ``auto_response_ssst`` function calls two functions that can be
used separatly. First, the function ``mask_for_response_ssst`` creates a mask
of voxels within the cuboid ROI and that meet the FA threshold constraint. This
mask can be used to calculate the number of voxels that were kept, or to also
apply an external mask (a WM mask for example). Second, the function
``response_from_mask_ssst`` takes the mask and returns the response function
calculated within the mask. If no changes are made to the mask between the to
calls, the resulting responses should be identical.
"""

mask = mask_for_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
nvoxels = np.sum(mask)
print(nvoxels)

response, ratio = response_from_mask_ssst(gtab, data, mask)

"""
The ``response`` tuple contains two elements. The first is an array with
the eigenvalues of the response function and the second is the average S0 for
this response.

It is good practice to always validate the result of auto_response_ssst. For
this purpose we can print the elements of ``response`` and have a look at their
values.
"""

print(response)

"""
(array([ 0.0014,  0.00029,  0.00029]), 416.206)

The tensor generated from the response must be prolate (two smaller eigenvalues
should be equal) and look anisotropic with a ratio of second to first
eigenvalue of about 0.2. Or in other words, the axial diffusivity of this
tensor should be around 5 times larger than the radial diffusivity.
"""

print(ratio)

"""
0.21197

We can double-check that we have a good response function by visualizing the
response function's ODF. Here is how you would do that:
"""

from dipy.viz import window, actor
from dipy.sims.voxel import single_tensor_odf

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
evals = response[0]
evecs = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]).T


response_odf = single_tensor_odf(default_sphere.vertices, evals, evecs)
# transform our data from 1D to 4D
response_odf = response_odf[None, None, None, :]
response_actor = actor.odf_slicer(response_odf, sphere=default_sphere,
                                  colormap='plasma')
scene.add(response_actor)
print('Saving illustration as csd_response.png')
window.record(scene, out_path='csd_response.png', size=(200, 200))
if interactive:
    window.show(scene)

"""
.. figure:: csd_response.png
   :align: center

   Estimated response function.

"""

scene.rm(response_actor)

"""
**Strategy 2 - data-driven calibration of response function** Depending
on the dataset, FA threshold may not be the best way to find the best possible
response function. For one, it depends on the diffusion tensor
(FA and first eigenvector), which has lower accuracy at high
b-values. Alternatively, the response function can be calibrated in a
data-driven manner [Tax2014]_.

First, the data is deconvolved with a 'fat' response function. All voxels that
are considered to contain only one peak in this deconvolution (as determined by
the peak threshold which gives an upper limit of the ratio of the second peak
to the first peak) are maintained, and from these voxels a new response
function is determined. This process is repeated until convergence is
reached. Here we calibrate the response function on a small part of the data.
"""

from dipy.reconst.csdeconv import recursive_response

"""
A WM mask can shorten computation time for the whole dataset. Here it is
created based on the DTI fit.
"""

import dipy.reconst.dti as dti
tenmodel = dti.TensorModel(gtab)
tenfit = tenmodel.fit(data, mask=data[..., 0] > 200)

from dipy.reconst.dti import fractional_anisotropy
FA = fractional_anisotropy(tenfit.evals)
MD = dti.mean_diffusivity(tenfit.evals)
wm_mask = (np.logical_or(FA >= 0.4, (np.logical_and(FA >= 0.15, MD >= 0.0011))))

response = recursive_response(gtab, data, mask=wm_mask, sh_order=8,
                              peak_thr=0.01, init_fa=0.08,
                              init_trace=0.0021, iter=8, convergence=0.001,
                              parallel=True)


"""
We can check the shape of the signal of the response function, which should be
like  a pancake:
"""

response_signal = response.on_sphere(default_sphere)
# transform our data from 1D to 4D
response_signal = response_signal[None, None, None, :]
response_actor = actor.odf_slicer(response_signal, sphere=default_sphere,
                                  colormap='plasma')

scene = window.Scene()

scene.add(response_actor)
print('Saving illustration as csd_recursive_response.png')
window.record(scene, out_path='csd_recursive_response.png', size=(200, 200))
if interactive:
    window.show(scene)

"""
.. figure:: csd_recursive_response.png
   :align: center

   Estimated response function using recursive calibration.

"""

scene.rm(response_actor)

"""
Step 2. fODF reconstruction
===========================

After estimating a response function for one of the strategies shown above,
we are ready to start the deconvolution process. Let's import the CSD model
and fit the datasets.
"""

from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel
csd_model = ConstrainedSphericalDeconvModel(gtab, response)

"""
For illustration purposes we will fit only a small portion of the data.
"""

data_small = data[20:50, 55:85, 38:39]
csd_fit = csd_model.fit(data_small)

"""
Show the CSD-based ODFs also known as FODFs (fiber ODFs).
"""

csd_odf = csd_fit.odf(default_sphere)

"""
Here we visualize only a 30x30 region.
"""

fodf_spheres = actor.odf_slicer(csd_odf, sphere=default_sphere, scale=0.9,
                                norm=False, colormap='plasma')

scene.add(fodf_spheres)

print('Saving illustration as csd_odfs.png')
window.record(scene, out_path='csd_odfs.png', size=(600, 600))
if interactive:
    window.show(scene)

"""
.. figure:: csd_odfs.png
   :align: center

   CSD ODFs.

In DIPY we also provide tools for finding the peak directions (maxima) of the
ODFs. For this purpose we recommend using ``peaks_from_model``.
"""

from dipy.direction import peaks_from_model

csd_peaks = peaks_from_model(model=csd_model,
                             data=data_small,
                             sphere=default_sphere,
                             relative_peak_threshold=.5,
                             min_separation_angle=25,
                             parallel=True)

scene.clear()
fodf_peaks = actor.peak_slicer(csd_peaks.peak_dirs, csd_peaks.peak_values)
scene.add(fodf_peaks)

print('Saving illustration as csd_peaks.png')
window.record(scene, out_path='csd_peaks.png', size=(600, 600))
if interactive:
    window.show(scene)

"""
.. figure:: csd_peaks.png
   :align: center

   CSD Peaks.

We can finally visualize both the ODFs and peaks in the same space.
"""

fodf_spheres.GetProperty().SetOpacity(0.4)

scene.add(fodf_spheres)

print('Saving illustration as csd_both.png')
window.record(scene, out_path='csd_both.png', size=(600, 600))
if interactive:
    window.show(scene)

"""
.. figure:: csd_both.png
   :align: center

   CSD Peaks and ODFs.

References
----------

.. [Tournier2007] J-D. Tournier, F. Calamante and A. Connelly, "Robust
   determination of the fibre orientation distribution in diffusion MRI:
   Non-negativity constrained super-resolved spherical deconvolution",
   Neuroimage, vol. 35, no. 4, pp. 1459-1472, 2007.

.. [Tax2014] C.M.W. Tax, B. Jeurissen, S.B. Vos, M.A. Viergever, A. Leemans,
   "Recursive calibration of the fiber response function for spherical
   deconvolution of diffusion MRI data", Neuroimage, vol. 86, pp. 67-80, 2014.

.. include:: ../links_names.inc

"""
